"""
rohmu - inotify wrapper

Copyright (c) 2016 Ohmu Ltd
See LICENSE for details
"""

from __future__ import annotations

from contextlib import suppress
from ctypes import c_char_p, c_int, c_uint32
from queue import Queue
from rohmu.typing import AnyPath
from threading import Thread
from typing import Iterable, Iterator, Optional

import ctypes
import logging
import os
import select
import struct


class InotifyEvent(ctypes.Structure):
    _fields_ = [
        ("wd", c_int),
        ("mask", c_uint32),
        ("cookie", c_uint32),
        ("len", c_uint32),
        ("name", c_char_p),
    ]


s_size = 16
# default 2048 events
INOTIFY_EVENT_BUFFER_SIZE = 2048 * (ctypes.sizeof(InotifyEvent) + s_size)

event_types = {
    "IN_CLOSE_WRITE": 0x00000008,
    "IN_CREATE": 0x00000100,
    "IN_MOVED_FROM": 0x00000040,
    "IN_MOVED_TO": 0x00000080,
    "IN_DELETE": 0x00000200,
    "IN_DELETE_SELF": 0x00000400,
    "IN_IGNORED": 0x8000,
}

IN_NONBLOCK = 0x00004000


def parse_inotify_buffer(event_buffer: bytes) -> Iterator[tuple[int, int, int, bytes]]:
    """Yield parsed inotify events from a raw bytes sequence.

    The `event_buffer` must be a byte sequence consisting of 0 or more sequences of the kind:

        +------------------+------+--------+-------------+------+
        | Watch descriptor | Mask | Cookie | Name length | Name |
        +------------------+------+--------+-------------+------+

    This function yields successive tuples (Watch descriptor, Mask, Cookie, Name) parsed from the provided bytes.

    """
    i = 0
    while i + s_size <= len(event_buffer):
        wd, mask, cookie, length = struct.unpack_from("iIII", event_buffer, i)
        name = event_buffer[i + s_size : i + s_size + length].rstrip(b"\0")
        i += s_size + length
        yield wd, mask, cookie, name


class InotifyWatcher(Thread):
    def __init__(self, compression_queue: Queue[dict[str, str]]) -> None:
        super().__init__()
        # use the newer form for future-proofness
        self.log = logging.getLogger("PGHoardInotify")
        self.libc = ctypes.CDLL("libc.so.6", use_errno=True)
        self.fd = self.libc.inotify_init()
        self.watch_to_path: dict[int, str] = {}
        self.cookies: dict[int, str] = {}
        self.running = True
        self.compression_queue = compression_queue
        self.timeout = 1.0
        self.log.debug("InotifyWatcher initialized")

    def add_watch(self, path: str, events: Optional[Iterable[str]] = None) -> None:
        mask = 0
        events = events or event_types.keys()
        for key in events:
            mask |= event_types[key]
        watch = self.libc.inotify_add_watch(self.fd, c_char_p(path.encode("utf8")), c_uint32(mask))
        if watch < 0:
            errno = ctypes.get_errno()
            raise OSError(errno, os.strerror(errno))
        self.watch_to_path[watch] = path
        self.log.debug("Added watch for path: %r", path)

    def read_events(self) -> None:
        event_buffer = None
        while self.running:
            with suppress(InterruptedError):
                rlist, _, _ = select.select([self.fd], [], [], self.timeout)
                if rlist:
                    for fd in rlist:
                        event_buffer = os.read(fd, INOTIFY_EVENT_BUFFER_SIZE)
                break
        if not event_buffer:
            return
        for wd, mask, cookie, name in parse_inotify_buffer(event_buffer):
            if wd == -1:
                continue
            self.create_event(wd, mask, cookie, name)

    def log_event(self, ev_type: str, full_path: AnyPath) -> None:
        if self.log.getEffectiveLevel() > logging.DEBUG:
            return

        try:
            st = os.stat(full_path)
        except:  # pylint: disable=bare-except
            st = None

        self.log.debug("event: %s %s, %r", full_path, ev_type, st)

    def create_event(self, wd: int, mask: int, cookie: int, name: bytes) -> None:
        if mask & event_types["IN_IGNORED"]:
            # explicit removal of watch or dir, ignore
            return

        decoded_name = name.decode("utf8")
        watched_path = self.watch_to_path[wd]
        full_path = os.path.join(self.watch_to_path[wd], decoded_name)

        if mask & event_types["IN_CREATE"] > 0:
            # file was created but writing to it is not finished yet
            self.log_event("IN_CREATE", full_path)
        elif mask & event_types["IN_CLOSE_WRITE"] > 0:
            # file was open for writing and was closed
            self.log_event("IN_CLOSE_WRITE", full_path)
            self.compression_queue.put({"type": "CLOSE_WRITE", "full_path": full_path, "watched_path": watched_path})
        elif mask & event_types["IN_DELETE"] > 0:
            self.log_event("IN_DELETE", full_path)
            self.compression_queue.put({"type": "DELETE", "full_path": full_path, "watched_path": watched_path})

        elif mask & event_types["IN_DELETE_SELF"] > 0:
            # the monitored directory was deleted
            self.log_event("IN_DELETE_SELF", full_path)
            directory = self.watch_to_path.pop(wd, "")
            self.log.debug("Directory: %r that we were watching has been deleted, removing watch", directory)
            self.libc.inotify_rm_watch(self.fd, wd)
        elif mask & event_types["IN_MOVED_FROM"] > 0:
            self.log_event("IN_MOVED_FROM", full_path)
            self.cookies[cookie] = full_path
        elif mask & event_types["IN_MOVED_TO"] > 0:
            self.log_event("IN_MOVED_TO", full_path)
            src_path = self.cookies.pop(cookie, None)
            if src_path:
                self.compression_queue.put(
                    {
                        "type": "MOVE",
                        "full_path": full_path,
                        "src_path": src_path,
                        "watched_path": watched_path,
                    }
                )
            else:
                self.compression_queue.put({"type": "CREATE", "full_path": full_path, "watched_path": watched_path})

    def run(self) -> None:
        self.log.debug("Starting InotifyWatcher")
        while self.running:
            self.read_events()
        self.log.debug("Quitting InotifyWatcher")
        os.close(self.fd)



from rohmu.compressor import CompressionFile, DecompressionFile

import io
import pytest


@pytest.mark.parametrize(
    "algorithm,contents",
    [
        ("lzma", b""),
        ("snappy", b""),
        ("lzma", b"Some contents"),
        ("snappy", b"Some contents"),
    ],
)
def test_compress_decompress_simple_file(algorithm: str, contents: bytes) -> None:
    bio = io.BytesIO()
    ef = CompressionFile(bio, algorithm=algorithm)
    ef.write(contents)
    ef.close()
    bio.seek(0)

    df = DecompressionFile(bio, algorithm=algorithm)
    data = df.read()
    assert data == contents



# Copyright (c) 2021 Aiven, Helsinki, Finland. https://aiven.io/

from io import BytesIO
from rohmu import rohmufile
from rohmu.errors import InvalidConfigurationError
from tempfile import NamedTemporaryFile
from typing import Any

import pytest


def test_fileobj_name(tmpdir: Any) -> None:
    with NamedTemporaryFile(dir=tmpdir, suffix="foo") as raw_output_obj:
        result = rohmufile._fileobj_name(raw_output_obj)  # type: ignore # pylint: disable=protected-access
        assert result.startswith("open file ")
        assert "foo" in result


def test_get_encryption_key_data_no_metadata() -> None:
    assert rohmufile._get_encryption_key_data(None, None) is None  # pylint: disable=protected-access
    assert rohmufile._get_encryption_key_data({}, None) is None  # pylint: disable=protected-access


def test_get_encryption_key_data_invalid_configuration() -> None:
    metadata = {"encryption-key-id": "foo"}
    with pytest.raises(InvalidConfigurationError, match="File is encrypted with key 'foo' but key not found"):
        rohmufile._get_encryption_key_data(metadata, lambda key_id: None)  # pylint: disable=protected-access


def test_get_encryption_key_data() -> None:
    def _getkey(key_id: str) -> str:
        assert key_id == "foo"
        return "bar"

    metadata = {"encryption-key-id": "foo"}
    key_data = rohmufile._get_encryption_key_data(metadata, _getkey)  # pylint: disable=protected-access
    assert key_data == "bar"


def test_file_reader_no_metadata() -> None:
    fileobj = BytesIO(b"foo")
    assert rohmufile.file_reader(fileobj=fileobj) == fileobj


def test_file_reader_no_key() -> None:
    fileobj = BytesIO(b"foo")
    metadata = {"encryption-key-id": "foo"}
    with pytest.raises(InvalidConfigurationError, match="File is encrypted with key 'foo' but key not found"):
        rohmufile.file_reader(fileobj=fileobj, metadata=metadata)


def test_file_reader_no_encryption_compression() -> None:
    fileobj = BytesIO(b"foo")
    metadata = {"a": "b"}
    assert rohmufile.file_reader(fileobj=fileobj, metadata=metadata) == fileobj



from datetime import datetime
from io import BytesIO
from pathlib import Path
from rohmu import errors
from rohmu.object_storage.local import LocalTransfer
from typing import Any

import pytest

DUMMY_CONTENT = b"dummy"
DUMMY_METADATA = {"Content-Length": str(len(DUMMY_CONTENT))}


@pytest.mark.parametrize("transfer_type", ["local_transfer"])
def test_nonexistent(transfer_type: str, request: Any) -> None:
    transfer = request.getfixturevalue(transfer_type)
    with pytest.raises(errors.FileNotFoundFromStorageError):
        transfer.get_metadata_for_key("NONEXISTENT")
    with pytest.raises(errors.FileNotFoundFromStorageError):
        transfer.delete_key("NONEXISTENT")
    with pytest.raises(errors.FileNotFoundFromStorageError):
        transfer.get_contents_to_file("NONEXISTENT", "nonexistent/a")
    with pytest.raises(errors.FileNotFoundFromStorageError):
        transfer.get_contents_to_fileobj("NONEXISTENT", BytesIO())
    with pytest.raises(errors.FileNotFoundFromStorageError):
        transfer.get_contents_to_string("NONEXISTENT")
    assert transfer.list_path("") == []
    assert transfer.list_path("NONEXISTENT") == []


@pytest.mark.parametrize("transfer_type", ["local_transfer"])
def test_basic_upload(transfer_type: str, tmp_path: Path, request: Any) -> None:
    scratch = tmp_path / "scratch"
    scratch.mkdir()
    transfer = request.getfixturevalue(transfer_type)
    sent_metadata = {"k": "v"}
    metadata = DUMMY_METADATA.copy()
    metadata.update(sent_metadata)
    transfer.store_file_from_memory("x1", DUMMY_CONTENT, sent_metadata)

    assert transfer.get_contents_to_string("x1") == (DUMMY_CONTENT, metadata)
    # Same thing, but with a key looking like a directory
    transfer.store_file_from_memory("NONEXISTENT-DIR/x1", DUMMY_CONTENT, None)
    assert transfer.get_contents_to_string("NONEXISTENT-DIR/x1") == (DUMMY_CONTENT, DUMMY_METADATA)

    # Same thing, but from disk now
    dummy_file = scratch / "a"
    with open(dummy_file, "wb") as fp:
        fp.write(DUMMY_CONTENT)
    transfer.store_file_from_disk("test1/x1", dummy_file, None)
    out = BytesIO()

    assert transfer.get_contents_to_fileobj("test1/x1", out) == DUMMY_METADATA
    assert out.getvalue() == DUMMY_CONTENT


@pytest.mark.parametrize("transfer_type", ["local_transfer"])
def test_copy(transfer_type: str, request: Any) -> None:
    transfer = request.getfixturevalue(transfer_type)
    sent_metadata = {"k": "v"}
    metadata = DUMMY_METADATA.copy()
    metadata.update(sent_metadata)
    transfer.store_file_from_memory("dummy", DUMMY_CONTENT, sent_metadata)
    transfer.copy_file(source_key="dummy", destination_key="dummy_copy")
    assert transfer.get_contents_to_string("dummy_copy") == (DUMMY_CONTENT, metadata)

    # Same thing, but with different metadata
    transfer.copy_file(source_key="dummy", destination_key="dummy_copy_metadata", metadata={"new_k": "new_v"})
    assert transfer.get_contents_to_string("dummy_copy_metadata") == (DUMMY_CONTENT, {"new_k": "new_v"})


@pytest.mark.parametrize("transfer_type", ["local_transfer"])
def test_list(transfer_type: str, request: Any) -> None:
    transfer = request.getfixturevalue(transfer_type)
    assert transfer.list_path("") == []

    # Test with a single file at root
    sent_metadata = {"k": "v"}
    metadata = DUMMY_METADATA.copy()
    metadata.update(sent_metadata)
    transfer.store_file_from_memory("dummy", DUMMY_CONTENT, metadata=sent_metadata)
    file_list = transfer.list_path("")
    assert len(file_list) == 1
    assert file_list[0]["name"] == "dummy"
    assert file_list[0]["metadata"] == metadata
    assert file_list[0]["size"] == len("dummy")
    assert isinstance(file_list[0]["last_modified"], datetime)

    # Test with a "subdirectory"
    transfer.store_file_from_memory("dummydir/dummy", DUMMY_CONTENT, metadata=sent_metadata)
    assert len(transfer.list_path("")) == 1
    assert set(transfer.iter_prefixes("")) == {"dummydir"}
    assert len(transfer.list_path("dummydir")) == 1

    files = transfer.list_path("", deep=True)
    assert len(files) == 2
    assert set(f["name"] for f in files) == {"dummy", "dummydir/dummy"}


def test_hidden_local_files(local_transfer: LocalTransfer) -> None:
    """
    Local storage specific test.
    """
    storage_dir = Path(local_transfer.prefix)
    # Since we've never used the local storage, need to create the directory
    storage_dir.mkdir()
    # When creating the file manually, we need to create some metadata too.
    with open(Path(local_transfer.prefix) / ".null", "w"):
        pass
    with open(Path(local_transfer.prefix) / ".null.metadata", "w") as f:
        f.write('{"k": "v", "_hash": ""}')
    assert local_transfer.list_path("") == []

    # Make sure the previous test actually worked, by manually creating a file
    # that must appear.
    with open(Path(local_transfer.prefix) / "somefile", "w"):
        pass
    with open(Path(local_transfer.prefix) / "somefile.metadata", "w") as f:
        f.write('{"k": "v", "_hash": ""}')

    files = local_transfer.list_path("")
    assert len(files) == 1
    assert files[0]["name"] == "somefile"


@pytest.mark.parametrize("transfer", ["local_transfer"])
def test_delete(transfer: LocalTransfer, request: Any) -> None:
    transfer = request.getfixturevalue(transfer)

    def setup() -> None:
        assert transfer.list_path("") == []
        transfer.store_file_from_memory("shallow", b"1")
        transfer.store_file_from_memory("something/quite/deep", b"2")
        assert len(transfer.list_path("", deep=True)) == 2

    setup()
    transfer.delete_tree("")

    setup()
    transfer.delete_keys(["shallow", "something/quite/deep"])

    setup()
    transfer.delete_key("shallow")
    transfer.delete_key("something/quite/deep")

    assert transfer.list_path("") == []



"""
rohmu - rohmu data transformation interface

Copyright (c) 2016 Ohmu Ltd
See LICENSE for details
"""

from __future__ import annotations

from . import IO_BLOCK_SIZE
from .compressor import CompressionFile, DecompressionFile, DecompressSink
from .encryptor import DecryptorFile, DecryptSink, EncryptorFile
from .errors import InvalidConfigurationError
from .filewrap import ThrottleSink
from .object_storage.base import IncrementalProgressCallbackType
from .typing import FileLike, HasWrite, Metadata
from contextlib import suppress
from inspect import signature
from typing import Callable, Optional, Union

import time


def _fileobj_name(input_obj: FileLike) -> str:
    if hasattr(input_obj, "name"):
        return "open file {!r}".format(getattr(input_obj, "name"))

    return repr(input_obj)


def _get_encryption_key_data(
    metadata: Optional[Metadata], key_lookup: Optional[Callable[[str], Optional[str]]]
) -> Optional[str]:
    if not metadata or not metadata.get("encryption-key-id"):
        return None

    key_id = metadata["encryption-key-id"]
    key_data = None
    if key_lookup:
        with suppress(KeyError):
            key_data = key_lookup(key_id)

    if not key_data:
        raise InvalidConfigurationError("File is encrypted with key {!r} but key not found".format(key_id))
    return key_data


def file_reader(
    *,
    fileobj: FileLike,
    metadata: Optional[Metadata] = None,
    key_lookup: Optional[Callable[[str], Optional[str]]] = None,
) -> FileLike:
    if not metadata:
        return fileobj

    key_data = _get_encryption_key_data(metadata, key_lookup)
    if key_data:
        fileobj = DecryptorFile(fileobj, key_data)

    comp_alg = metadata.get("compression-algorithm")
    if comp_alg:
        fileobj = DecompressionFile(fileobj, comp_alg)

    return fileobj


def create_sink_pipeline(
    *,
    output: HasWrite,
    file_size: int = 0,
    metadata: Optional[Metadata] = None,
    key_lookup: Optional[Callable[[str], Optional[str]]] = None,
    throttle_time: float = 0.001,
) -> HasWrite:
    if throttle_time:
        output = ThrottleSink(output, throttle_time)

    comp_alg = metadata.get("compression-algorithm") if metadata else None
    if comp_alg:
        output = DecompressSink(output, comp_alg)

    key_data = _get_encryption_key_data(metadata, key_lookup)
    if key_data:
        output = DecryptSink(output, file_size, key_data)

    return output


def _callback_wrapper(progress_callback: IncrementalProgressCallbackType) -> IncrementalProgressCallbackType:
    # Gracefully support legacy callbacks which do not expect any arguments to be passed to them
    if progress_callback is None:
        return None
    sig = signature(progress_callback)
    if len(sig.parameters) == 0:
        return lambda f: progress_callback()  # type: ignore [misc,call-arg]
    return progress_callback


def read_file(
    *,
    input_obj: FileLike,
    output_obj: FileLike,
    metadata: Metadata,
    key_lookup: Optional[Callable[[str], Optional[str]]],
    progress_callback: IncrementalProgressCallbackType = None,
    log_func: Optional[Callable[..., None]] = None,
) -> tuple[int, int]:
    start_time = time.monotonic()
    progress_callback = _callback_wrapper(progress_callback)

    with file_reader(fileobj=input_obj, metadata=metadata, key_lookup=key_lookup) as fp_in:
        while True:
            input_data = fp_in.read(IO_BLOCK_SIZE)
            if not input_data:
                break

            output_obj.write(input_data)
            if progress_callback:
                progress_callback(len(input_data))

    original_size = input_obj.tell()
    result_size = output_obj.tell()

    if log_func:
        action = "Decompressed"
        if metadata.get("encryption-key-id"):
            action += " and decrypted"

        log_func(
            "%s %d bytes to %d bytes in %s, took: %.3fs",
            action,
            original_size,
            result_size,
            _fileobj_name(output_obj),
            time.monotonic() - start_time,
        )

    return original_size, result_size


def file_writer(
    *,
    fileobj: FileLike,
    compression_algorithm: Optional[str] = None,
    compression_level: int = 0,
    compression_threads: int = 0,
    rsa_public_key: Union[None, str, bytes] = None,
) -> FileLike:
    if rsa_public_key:
        fileobj = EncryptorFile(fileobj, rsa_public_key)

    if compression_algorithm:
        fileobj = CompressionFile(fileobj, compression_algorithm, compression_level, compression_threads)

    return fileobj


def write_file(
    *,
    input_obj: FileLike,
    output_obj: FileLike,
    progress_callback: IncrementalProgressCallbackType = None,
    compression_algorithm: Optional[str] = None,
    compression_level: int = 0,
    compression_threads: int = 0,
    rsa_public_key: Union[None, str, bytes] = None,
    log_func: Optional[Callable[..., None]] = None,
    header_func: Optional[Callable[[bytes], None]] = None,
    data_callback: Optional[Callable[[bytes], None]] = None,
) -> tuple[int, int]:
    start_time = time.monotonic()
    progress_callback = _callback_wrapper(progress_callback)

    original_size = 0
    with file_writer(
        fileobj=output_obj,
        compression_algorithm=compression_algorithm,
        compression_threads=compression_threads,
        compression_level=compression_level,
        rsa_public_key=rsa_public_key,
    ) as fp_out:
        header_block = True
        while True:
            input_data = input_obj.read(IO_BLOCK_SIZE)
            if not input_data:
                break

            if data_callback:
                data_callback(input_data)

            if header_block and header_func:
                header_func(input_data)
                header_block = False

            fp_out.write(input_data)
            original_size += len(input_data)
            if progress_callback:
                progress_callback(len(input_data))

    result_size = output_obj.tell()

    if log_func:
        log_compression_result(
            elapsed=time.monotonic() - start_time,
            encrypted=bool(rsa_public_key),
            log_func=log_func,
            original_size=original_size,
            result_size=result_size,
            source_name=_fileobj_name(input_obj),
        )

    return original_size, result_size


def log_compression_result(
    *, log_func: Callable[..., None], source_name: str, original_size: int, result_size: int, encrypted: bool, elapsed: float
) -> None:
    if original_size <= result_size:
        action = "Stored"
        ratio = ""
    else:
        action = "Compressed"
        ratio = " ({:.0%})".format(result_size / original_size)

    if encrypted:
        action += " and encrypted"

    log_func("%s %d byte of %s to %d bytes%s, took: %.3fs", action, original_size, source_name, result_size, ratio, elapsed)



"""Copyright (c) 2023 Aiven, Helsinki, Finland. https://aiven.io/"""
from dataclasses import dataclass
from rohmu import transfer_pool
from rohmu.errors import InvalidTransferError
from rohmu.object_storage.local import LocalTransfer
from unittest.mock import patch

import pytest
import time


@dataclass(frozen=True)
class FakeTransfer:
    name: str
    is_thread_safe: bool = False
    # needed for safe transfer tests
    prefix: str = "prefix"
    notifier: str = "notifier"
    stats: str = "stats"
    config_model: str = "config_model"


def test_transfer_pool() -> None:
    """Test that a cached transfer object is returned, if available."""
    storage_config_1 = {"storage_type": "azure"}
    storage_config_2 = {"storage_type": "gcp"}

    with patch("rohmu.transfer_pool.rohmu_get_transfer") as rohmu_get_transfer:
        pool = transfer_pool.TransferPool()
        rohmu_get_transfer.return_value = FakeTransfer("mock_transfer_1")
        with pool.with_transfer(storage_config_1) as transfer:
            rohmu_get_transfer.assert_called_once_with(storage_config_1)
            assert transfer is not None

        rohmu_get_transfer.return_value = FakeTransfer("mock_transfer_2")
        with pool.with_transfer(storage_config_1) as next_transfer:
            assert next_transfer == transfer
            with pool.with_transfer(storage_config_2) as next_transfer_2:
                assert next_transfer_2 is not None
                assert next_transfer_2 != transfer


def test_transfer_pool_context_manager_returns_transfer_on_exception() -> None:
    """Test that a cached transfer object is returned, if available."""
    storage_config_1 = {"storage_type": "azure"}

    with patch("rohmu.transfer_pool.rohmu_get_transfer") as rohmu_get_transfer:
        pool = transfer_pool.TransferPool()
        rohmu_get_transfer.return_value = FakeTransfer("mock_transfer_1")
        try:
            with pool.with_transfer(storage_config_1) as transfer:
                raise ValueError("Some error happens")
        except ValueError:
            pass

        rohmu_get_transfer.return_value = FakeTransfer("mock_transfer_2")
        with pool.with_transfer(storage_config_1) as next_transfer:
            # the contextmanager should have returned the transfer to the pool, so we expect to get the same one
            assert next_transfer == transfer


def test_transfer_pool_expiration() -> None:
    """Test that a cached transfer object is not returned, if too old."""
    storage_config = {"storage_type": "azure"}

    with patch("rohmu.transfer_pool.rohmu_get_transfer") as rohmu_get_transfer:
        rohmu_get_transfer.return_value = FakeTransfer("mock_transfer_1")
        pool = transfer_pool.TransferPool()

        with pool.with_transfer(storage_config) as transfer:
            # let's create an item
            pass

        rohmu_get_transfer.return_value = FakeTransfer("mock_transfer_2")
        transfer_item = list(pool._caches.values())[0].transfers_heap[0]  # pylint: disable=protected-access
        transfer_item.create_time = time.monotonic() - (3600 * 2)
        with pool.with_transfer(storage_config) as next_transfer:
            assert next_transfer == transfer
        transfer_item.create_time = time.monotonic() - (3600 * 3)
        with pool.with_transfer(storage_config) as next_transfer:
            assert next_transfer != transfer


def test_transfer_pool_get_oldest() -> None:
    """Test that when getting a transfer object from a pool, the oldest one available is returned."""
    storage_config = {"storage_type": "azure"}

    with patch("rohmu.transfer_pool.rohmu_get_transfer") as rohmu_get_transfer:
        rohmu_get_transfer.return_value = FakeTransfer("mock_transfer_1")
        pool = transfer_pool.TransferPool()

        with pool.with_transfer(storage_config) as transfer_1:
            rohmu_get_transfer.return_value = FakeTransfer("mock_transfer_2")
            with pool.with_transfer(storage_config):
                # do nothing and put the transfers back in the pool
                pass

        with pool.with_transfer(storage_config) as next_transfer:
            assert next_transfer == transfer_1


def test_transfer_pool_for_thread_safe_transfer_always_returns_the_same_instance() -> None:
    """Test that when using a threadsafe transfer object the pool always returns the same instance"""
    storage_config = {"storage_type": "azure"}

    with patch("rohmu.transfer_pool.rohmu_get_transfer") as rohmu_get_transfer:
        rohmu_get_transfer.return_value = FakeTransfer("mock_transfer_1", is_thread_safe=True)
        pool = transfer_pool.TransferPool()

        with pool.with_transfer(storage_config) as transfer_1, pool.with_transfer(storage_config) as transfer_2:
            assert transfer_1 == transfer_2


def test_transfer_pool_for_thread_safe_transfer_does_not_expire() -> None:
    """Test that a cached transfer object is not returned, if too old, even for thread-safe transfers."""
    storage_config = {"storage_type": "azure"}

    with patch("rohmu.transfer_pool.rohmu_get_transfer") as rohmu_get_transfer:
        rohmu_get_transfer.return_value = FakeTransfer("mock_transfer_1", is_thread_safe=True)
        pool = transfer_pool.TransferPool()

        with pool.with_transfer(storage_config) as transfer:
            # let's create an item
            pass

        rohmu_get_transfer.return_value = FakeTransfer("mock_transfer_2", is_thread_safe=True)
        transfer_item = list(pool._caches.values())[0].transfers_heap[0]  # pylint: disable=protected-access
        transfer_item.create_time = time.monotonic() - (3600 * 2)
        with pool.with_transfer(storage_config) as next_transfer:
            assert next_transfer == transfer
        transfer_item.create_time = time.monotonic() - (3600 * 3)
        with pool.with_transfer(storage_config) as next_transfer:
            # for thread-safe we do not expire the transfers
            assert next_transfer == transfer


# pylint: disable=protected-access
def test_transfer_pool_get_return_api() -> None:
    """Test that a cached transfer object is returned, if available."""
    storage_config_1 = {"storage_type": "azure"}
    storage_config_2 = {"storage_type": "gcp"}

    with patch("rohmu.transfer_pool.rohmu_get_transfer") as rohmu_get_transfer:
        pool = transfer_pool.TransferPool()
        rohmu_get_transfer.return_value = FakeTransfer("mock_transfer_1")
        safe_transfer = pool.get_transfer(storage_config_1)
        transfer = safe_transfer._item.transfer
        rohmu_get_transfer.assert_called_once_with(storage_config_1)
        assert transfer is not None
        safe_transfer.return_to_pool()

        rohmu_get_transfer.return_value = FakeTransfer("mock_transfer_2")
        next_safe_transfer_1 = pool.get_transfer(storage_config_1)
        assert next_safe_transfer_1._item.transfer == transfer
        next_safe_transfer_2 = pool.get_transfer(storage_config_2)
        assert next_safe_transfer_2 is not None
        assert next_safe_transfer_2._item.transfer != transfer


# pylint: disable=protected-access
def test_transfer_pool_expiration_get_return_api() -> None:
    """Test that a cached transfer object is not returned, if too old."""
    storage_config = {"storage_type": "azure"}

    with patch("rohmu.transfer_pool.rohmu_get_transfer") as rohmu_get_transfer:
        rohmu_get_transfer.return_value = FakeTransfer("mock_transfer_1")
        pool = transfer_pool.TransferPool()

        safe_transfer = pool.get_transfer(storage_config)
        transfer_item = safe_transfer._item
        safe_transfer.return_to_pool()
        rohmu_get_transfer.return_value = FakeTransfer("mock_transfer_2")
        transfer_item.create_time = time.monotonic() - (3600 * 2)
        next_safe_transfer = pool.get_transfer(storage_config)
        assert next_safe_transfer._item.transfer == transfer_item.transfer
        next_safe_transfer.return_to_pool()
        transfer_item.create_time = time.monotonic() - (3600 * 3)
        next_safe_transfer = pool.get_transfer(storage_config)
        assert next_safe_transfer._item.transfer != transfer_item.transfer


# pylint: disable=protected-access
def test_transfer_pool_get_oldest_get_return_api() -> None:
    """Test that when getting a transfer object from a pool, the oldest one available is returned."""
    storage_config = {"storage_type": "azure"}

    with patch("rohmu.transfer_pool.rohmu_get_transfer") as rohmu_get_transfer:
        rohmu_get_transfer.return_value = FakeTransfer("mock_transfer_1")
        pool = transfer_pool.TransferPool()

        safe_transfer_1 = pool.get_transfer(storage_config)
        transfer_1 = safe_transfer_1._item.transfer
        rohmu_get_transfer.return_value = FakeTransfer("mock_transfer_2")
        safe_transfer_2 = pool.get_transfer(storage_config)
        safe_transfer_1.return_to_pool()
        safe_transfer_2.return_to_pool()

        next_safe_transfer = pool.get_transfer(storage_config)
        assert next_safe_transfer._item.transfer == transfer_1


# pylint: disable=protected-access
def test_transfer_pool_for_thread_safe_transfer_always_returns_the_same_instance_get_return_api() -> None:
    """Test that when using a threadsafe transfer object the pool always returns the same instance"""
    storage_config = {"storage_type": "azure"}

    with patch("rohmu.transfer_pool.rohmu_get_transfer") as rohmu_get_transfer:
        rohmu_get_transfer.return_value = FakeTransfer("mock_transfer_1", is_thread_safe=True)
        pool = transfer_pool.TransferPool()

        transfer_1 = pool.get_transfer(storage_config)
        transfer_2 = pool.get_transfer(storage_config)
        assert transfer_1._item.transfer == transfer_2._item.transfer


# pylint: disable=protected-access
def test_transfer_pool_for_thread_safe_transfer_does_not_expire_get_return_api() -> None:
    """Test that a cached transfer object is not returned, if too old, even for thread-safe transfers."""
    storage_config = {"storage_type": "azure"}

    with patch("rohmu.transfer_pool.rohmu_get_transfer") as rohmu_get_transfer:
        rohmu_get_transfer.return_value = FakeTransfer("mock_transfer_1", is_thread_safe=True)
        pool = transfer_pool.TransferPool()

        safe_transfer = pool.get_transfer(storage_config)
        transfer_item = safe_transfer._item
        transfer = transfer_item.transfer
        safe_transfer.return_to_pool()

        rohmu_get_transfer.return_value = FakeTransfer("mock_transfer_2", is_thread_safe=True)

        transfer_item.create_time = time.monotonic() - (3600 * 2)
        next_safe_transfer = pool.get_transfer(storage_config)
        assert next_safe_transfer._item.transfer == transfer
        next_safe_transfer.return_to_pool()
        transfer_item.create_time = time.monotonic() - (3600 * 3)
        next_safe_transfer = pool.get_transfer(storage_config)
        # We do not expire transfers for thread-safe transfers
        assert next_safe_transfer._item.transfer == transfer


@pytest.mark.parametrize(
    "attr_name",
    [
        "is_thread_safe",
        "config_model",
        "prefix",
        "notifier",
        "stats",
        "copy_file",
        "format_key_for_backend",
        "format_key_from_backend",
        "delete_key",
        "delete_keys",
        "delete_tree",
        "get_contents_to_file",
        "get_contents_to_fileobj",
        "get_contents_to_string",
        "get_file_size",
        "get_metadata_for_key",
        "list_path",
        "list_iter",
        "list_prefixes",
        "iter_prefixes",
        "iter_key",
        "sanitize_metadata",
        "store_file_from_memory",
        "store_file_from_disk",
        "store_file_object",
        "from_model",
    ],
)
def test_safe_transfer_prevents_access_after_returning(attr_name: str) -> None:
    """Test that a cached transfer object is not returned, if too old, even for thread-safe transfers."""
    storage_config = {"storage_type": "azure"}

    with patch("rohmu.transfer_pool.rohmu_get_transfer") as rohmu_get_transfer:
        rohmu_get_transfer.return_value = LocalTransfer("mock_transfer_1")
        pool = transfer_pool.TransferPool()

        safe_transfer = pool.get_transfer(storage_config)
        getattr(safe_transfer, attr_name)
        safe_transfer.return_to_pool()
        with pytest.raises(InvalidTransferError):
            getattr(safe_transfer, attr_name)



"""
rohmu - openstack swift object store interface

Copyright (c) 2016 Ohmu Ltd
Copyright (c) 2022 Aiven, Helsinki, Finland. https://aiven.io/
See LICENSE for details
"""

from __future__ import annotations

from ..common.models import StorageModel
from ..common.statsd import StatsdConfig
from ..dates import parse_timestamp
from ..errors import FileNotFoundFromStorageError
from ..notifier.interface import Notifier
from ..typing import Metadata
from .base import (
    BaseTransfer,
    IncrementalProgressCallbackType,
    IterKeyItem,
    KEY_TYPE_OBJECT,
    KEY_TYPE_PREFIX,
    ProgressProportionCallbackType,
)
from contextlib import suppress
from swiftclient import client, exceptions  # pylint: disable=import-error
from typing import Any, BinaryIO, Iterator, Optional, Tuple

import logging
import os
import time

CHUNK_SIZE = 1024 * 1024 * 5  # 5 Mi
SEGMENT_SIZE = 1024 * 1024 * 1024 * 3  # 3 Gi


# Swift client logs excessively at INFO level, outputting things like a curl
# command line to recreate the request that failed with a simple 404 error.
# At WARNING level curl commands are not logged, but we get a full ugly
# traceback for all failures, including 404s.  Monkey-patch them away.
def swift_exception_logger(err: BaseException) -> Any:
    if not isinstance(err, exceptions.ClientException):
        return orig_swift_exception_logger(err)
    if getattr(err, "http_status", None) is None:
        return orig_swift_exception_logger(err)
    if err.http_status == 404 and err.msg.startswith("Object GET failed"):
        client.logger.debug("GET %r FAILED: %r", err.http_path, err.http_status)
    else:
        client.logger.error(str(err))
    return None


orig_swift_exception_logger = client.logger.exception
client.logger.exception = swift_exception_logger
logging.getLogger("swiftclient").setLevel(logging.WARNING)


class Config(StorageModel):
    user: str
    key: str
    container_name: str
    auth_url: str
    auth_version: str = "2.0"
    tenant_name: Optional[str] = None
    segment_size: int = SEGMENT_SIZE
    region_name: Optional[str] = None
    user_id: Optional[str] = None
    user_domain_id: Optional[str] = None
    user_domain_name: Optional[str] = None
    tenant_id: Optional[str] = None
    project_id: Optional[str] = None
    project_name: Optional[str] = None
    project_domain_id: Optional[str] = None
    project_domain_name: Optional[str] = None
    service_type: Optional[str] = None
    endpoint_type: Optional[str] = None


class SwiftTransfer(BaseTransfer[Config]):
    config_model = Config

    def __init__(
        self,
        *,
        user: str,
        key: str,
        container_name: str,
        auth_url: str,
        auth_version: str = "2.0",
        tenant_name: Optional[str] = None,
        prefix: Optional[str] = None,
        segment_size: int = SEGMENT_SIZE,
        region_name: Optional[str] = None,
        user_id: Optional[str] = None,
        user_domain_id: Optional[str] = None,
        user_domain_name: Optional[str] = None,
        tenant_id: Optional[str] = None,
        project_id: Optional[str] = None,
        project_name: Optional[str] = None,
        project_domain_id: Optional[str] = None,
        project_domain_name: Optional[str] = None,
        service_type: Optional[str] = None,
        endpoint_type: Optional[str] = None,
        notifier: Optional[Notifier] = None,
        statsd_info: Optional[StatsdConfig] = None,
    ) -> None:
        prefix = prefix.lstrip("/") if prefix else ""
        super().__init__(prefix=prefix, notifier=notifier, statsd_info=statsd_info)
        self.container_name = container_name

        if auth_version == "3.0":
            os_options = {
                "region_name": region_name,
                "user_id": user_id,
                "user_domain_id": user_domain_id,
                "user_domain_name": user_domain_name,
                "tenant_id": tenant_id,
                "project_id": project_id,
                "project_name": project_name,
                "project_domain_id": project_domain_id,
                "project_domain_name": project_domain_name,
                "service_type": service_type,
                "endpoint_type": endpoint_type,
            }
        else:
            if region_name is not None:
                os_options = {"region_name": region_name}
            else:
                os_options = None

        self.conn = client.Connection(
            user=user, key=key, authurl=auth_url, tenant_name=tenant_name, auth_version=auth_version, os_options=os_options
        )
        self.container = self.get_or_create_container(self.container_name)
        self.segment_size = segment_size
        self.log.debug("SwiftTransfer initialized")

    @staticmethod
    def _headers_to_metadata(headers: dict[str, str]) -> Metadata:
        return {name[len("x-object-meta-") :]: value for name, value in headers.items() if name.startswith("x-object-meta-")}

    @staticmethod
    def _metadata_to_headers(metadata: Metadata) -> dict[str, str]:
        return {"x-object-meta-{}".format(name): str(value) for name, value in metadata.items()}

    def get_metadata_for_key(self, key: str) -> Metadata:
        path = self.format_key_for_backend(key)
        return self._metadata_for_key(path)

    def _metadata_for_key(self, key: str, *, resolve_manifest: bool = False) -> Metadata:
        try:
            headers = self.conn.head_object(self.container_name, key)
        except exceptions.ClientException as ex:
            if ex.http_status == 404:
                raise FileNotFoundFromStorageError(key)
            raise

        metadata = self._headers_to_metadata(headers)

        if resolve_manifest and "x-object-manifest" in headers:
            manifest = headers["x-object-manifest"]
            seg_container, seg_prefix = manifest.split("/", 1)
            _, segments = self.conn.get_container(seg_container, prefix=seg_prefix, delimiter="/")
            segments_size = sum(item["bytes"] for item in segments if "bytes" in item)
            metadata["_segments_size"] = segments_size

        return metadata

    def iter_key(
        self, key: str, *, with_metadata: bool = True, deep: bool = False, include_key: bool = False
    ) -> Iterator[IterKeyItem]:
        path = self.format_key_for_backend(key, remove_slash_prefix=True, trailing_slash=not include_key)
        self.log.debug("Listing path %r", path)
        if not deep:
            kwargs = {"delimiter": "/"}
        else:
            kwargs = {}
        _, results = self.conn.get_container(self.container_name, prefix=path, full_listing=True, **kwargs)
        for item in results:
            if "subdir" in item:
                yield IterKeyItem(type=KEY_TYPE_PREFIX, value=self.format_key_from_backend(item["subdir"]).rstrip("/"))
            else:
                if with_metadata:
                    metadata = self._metadata_for_key(item["name"], resolve_manifest=True)
                    segments_size = metadata.pop("_segments_size", 0)
                else:
                    metadata = None
                    segments_size = 0
                last_modified = parse_timestamp(item["last_modified"])

                # Response format is documented at:
                #
                #   https://docs.openstack.org/api-ref/object-store/?expanded=list-endpoints-detail,show-container-details-and-list-objects-detail # pylint: disable=line-too-long
                yield IterKeyItem(
                    type=KEY_TYPE_OBJECT,
                    value={
                        "name": self.format_key_from_backend(item["name"]),
                        "size": item["bytes"] + segments_size,
                        "last_modified": last_modified,
                        "metadata": metadata,
                        "hash": item["hash"],
                    },
                )

    def _delete_object_plain(self, key: str) -> None:
        try:
            return self.conn.delete_object(self.container_name, key)
        except exceptions.ClientException as ex:
            if ex.http_status == 404:
                raise FileNotFoundFromStorageError(key)
            raise

    def _delete_object_segments(self, key: str, manifest: str) -> None:
        self._delete_object_plain(key)
        seg_container, seg_prefix = manifest.split("/", 1)
        _, segments = self.conn.get_container(seg_container, prefix=seg_prefix, delimiter="/")
        for item in segments:
            if "name" in item:
                with suppress(FileNotFoundFromStorageError):
                    self._delete_object_plain(item["name"])

    def delete_key(self, key: str) -> None:
        path = self.format_key_for_backend(key)
        self.log.debug("Deleting key: %r", path)
        try:
            headers = self.conn.head_object(self.container_name, path)
        except exceptions.ClientException as ex:
            if ex.http_status == 404:
                raise FileNotFoundFromStorageError(path)
            raise
        if "x-object-manifest" in headers:
            self._delete_object_segments(path, headers["x-object-manifest"])
        else:
            self._delete_object_plain(path)
        self.notifier.object_deleted(key=key)

    def get_contents_to_fileobj(
        self,
        key: str,
        fileobj_to_store_to: BinaryIO,
        *,
        byte_range: Optional[Tuple[int, int]] = None,
        progress_callback: ProgressProportionCallbackType = None,
    ) -> Metadata:
        if byte_range:
            # TODO if someday relevant. swift API itself implements it,
            # c.f. https://docs.openstack.org/api-ref/object-store/
            raise NotImplementedError("byte range fetching not supported")
        path = self.format_key_for_backend(key)
        try:
            headers, data_gen = self.conn.get_object(self.container_name, path, resp_chunk_size=CHUNK_SIZE)
        except exceptions.ClientException as ex:
            if ex.http_status == 404:
                raise FileNotFoundFromStorageError(path)
            raise

        content_len = int(headers.get("content-length") or 0)
        current_pos = 0
        for chunk in data_gen:
            fileobj_to_store_to.write(chunk)
            if progress_callback:
                if not content_len:
                    # if content length is not known we'll always say we're half-way there
                    progress_callback(1, 2)
                else:
                    current_pos += len(chunk)
                    progress_callback(current_pos, content_len)

        return self._headers_to_metadata(headers)

    def get_file_size(self, key: str) -> int:
        # Not implemented due to lack of environment where to test this. This method is not required by
        # PGHoard itself, this is only called by external apps that utilize PGHoard's object storage abstraction.
        raise NotImplementedError

    def get_or_create_container(self, container_name: str) -> str:
        start_time = time.monotonic()
        try:
            self.conn.get_container(container_name, headers={}, limit=1)  # Limit 1 here to not traverse the entire folder
        except exceptions.ClientException as ex:
            if ex.http_status == 404:
                self.conn.put_container(container_name, headers={})
                self.log.debug(
                    "Created container: %r successfully, took: %.3fs",
                    container_name,
                    time.monotonic() - start_time,
                )
                return container_name
            raise
        return container_name

    def copy_file(
        self, *, source_key: str, destination_key: str, metadata: Optional[Metadata] = None, **_kwargs: Any
    ) -> None:
        source_key = self.format_key_for_backend(source_key)
        destination_key = "/".join((self.container_name, self.format_key_for_backend(destination_key)))
        sanitized_metadata = self.sanitize_metadata(metadata)
        headers: Metadata = self._metadata_to_headers(sanitized_metadata)
        if metadata:
            headers["X-Fresh-Metadata"] = True
        self.conn.copy_object(self.container_name, source_key, destination=destination_key, headers=headers)
        self.notifier.object_copied(key=destination_key, size=None, metadata=sanitized_metadata)

    def store_file_object(
        self,
        key: str,
        fd: BinaryIO,
        metadata: Optional[Metadata] = None,
        *,
        cache_control: Optional[str] = None,
        mimetype: Optional[str] = None,
        multipart: Optional[bool] = None,
        upload_progress_fn: IncrementalProgressCallbackType = None,
    ) -> None:  # pylint: disable=unused-argument
        metadata = metadata or {}
        content_length = metadata.get("Content-Length")
        multipart = self._should_multipart(
            chunk_size=self.segment_size, default=True, metadata=metadata, multipart=multipart
        )
        self._store_file_contents(
            key,
            fd,
            cache_control=cache_control,
            metadata=metadata,
            mimetype=mimetype,
            upload_progress_fn=upload_progress_fn,
            multipart=multipart,
            content_length=content_length,
        )
        self.notifier.object_created(key=key, size=content_length, metadata=self.sanitize_metadata(metadata))

    def _store_file_contents(
        self,
        key: str,
        fp: BinaryIO,
        cache_control: Optional[str] = None,
        metadata: Optional[Metadata] = None,
        mimetype: Optional[str] = None,
        upload_progress_fn: IncrementalProgressCallbackType = None,
        multipart: Optional[bool] = None,
        content_length: Optional[int] = None,
    ) -> None:
        if cache_control is not None:
            raise NotImplementedError("SwiftTransfer: cache_control support not implemented")

        if multipart:
            # Start by trying to delete the file - if it's a potential multipart file we need to manually
            # delete it, otherwise old segments won't be cleaned up by anything.
            # chunks.
            with suppress(FileNotFoundFromStorageError):
                self.delete_key(key)
        path = self.format_key_for_backend(key)
        headers = self._metadata_to_headers(self.sanitize_metadata(metadata))
        # Fall back to the "one segment" if possible
        if (not multipart) or (not content_length) or content_length <= self.segment_size:
            self.log.debug("Uploading %r to %r (%r bytes)", fp, path, content_length)
            self.conn.put_object(self.container_name, path, contents=fp, content_length=content_length, headers=headers)
            if upload_progress_fn and content_length is not None:
                upload_progress_fn(content_length)
            return

        # Segmented transfer
        # upload segments of a file like `backup-bucket/site-name/basebackup/2016-03-22_0`
        # to as `backup-bucket/site-name/basebackup_segments/2016-03-22_0/{:08x}`
        segment_no = 0
        segment_path = "{}_segments/{}/".format(os.path.dirname(path), os.path.basename(path))
        segment_key_format = "{}{{:08x}}".format(segment_path).format
        remaining = content_length
        while remaining > 0:
            this_segment_size = min(self.segment_size, remaining)
            remaining -= this_segment_size
            segment_no += 1
            self.log.debug("Uploading segment %r of %r to %r (%r bytes)", segment_no, fp, path, this_segment_size)
            segment_key = segment_key_format(segment_no)  # pylint: disable=too-many-format-args
            self.conn.put_object(
                self.container_name, segment_key, contents=fp, content_length=this_segment_size, content_type=mimetype
            )
            if upload_progress_fn:
                upload_progress_fn(content_length - remaining)
        self.log.info("Uploaded %r segments of %r to %r", segment_no, path, segment_path)
        headers["x-object-manifest"] = "{}/{}".format(self.container_name, segment_path.lstrip("/"))
        self.conn.put_object(self.container_name, path, contents="", headers=headers, content_length=0)



from __future__ import annotations

from rohmu.inotify import parse_inotify_buffer

import pytest
import struct


def _make_buffer(wd: int, mask: int, cookie: int, name: bytes) -> bytes:
    return struct.pack("iIII", wd, mask, cookie, len(name)) + name


@pytest.mark.parametrize(
    "buffer,result",
    [
        (b"", []),
        (_make_buffer(1, 2, 3, b"Hello, World!"), [(1, 2, 3, b"Hello, World!")]),
        (
            _make_buffer(11, 12, 13, b"Hello, ") + _make_buffer(101, 102, 103, b"World!"),
            [(11, 12, 13, b"Hello, "), (101, 102, 103, b"World!")],
        ),
    ],
)
def test_parse_inotify_buffer(buffer: bytes, result: list[tuple[int, int, int, bytes]]) -> None:
    got = list(parse_inotify_buffer(buffer))
    assert got == result



"""
rohmu - google cloud object store interface

Copyright (c) 2016 Ohmu Ltd
Copyright (c) 2022 Aiven, Helsinki, Finland. https://aiven.io/
See LICENSE for details
"""

from __future__ import annotations

from ..common.models import ProxyInfo, StorageModel, StorageOperation
from ..common.statsd import StatsClient, StatsdConfig
from ..dates import parse_timestamp
from ..errors import FileNotFoundFromStorageError, InvalidByteRangeError, InvalidConfigurationError
from ..notifier.interface import Notifier
from ..typing import AnyPath, Metadata
from ..util import get_total_size_from_content_range
from .base import (
    BaseTransfer,
    get_total_memory,
    IncrementalProgressCallbackType,
    IterKeyItem,
    KEY_TYPE_OBJECT,
    KEY_TYPE_PREFIX,
    ProgressProportionCallbackType,
)
from contextlib import contextmanager
from googleapiclient.discovery import build, Resource
from googleapiclient.errors import HttpError
from googleapiclient.http import (
    build_http,
    HttpRequest,
    MediaDownloadProgress,
    MediaIoBaseDownload,
    MediaUpload,
    MediaUploadProgress,
)
from http.client import IncompleteRead
from oauth2client import GOOGLE_TOKEN_URI
from oauth2client.client import GoogleCredentials
from typing import Any, BinaryIO, Callable, Dict, Iterable, Iterator, Optional, TextIO, Tuple, TypeVar, Union
from typing_extensions import Protocol

import codecs
import dataclasses
import errno

# NOTE: this import is not needed per-se, but it's imported here first to point the
# user to the most important possible missing dependency
import googleapiclient  # noqa pylint: disable=unused-import
import httplib2
import json
import logging
import os
import random
import socket
import ssl
import time

try:
    from oauth2client.service_account import ServiceAccountCredentials

    ServiceAccountCredentials_from_dict = ServiceAccountCredentials.from_json_keyfile_dict
except ImportError:
    from oauth2client.service_account import _ServiceAccountCredentials

    def ServiceAccountCredentials_from_dict(
        credentials: dict[str, Any], scopes: Optional[list[str]] = None
    ) -> GoogleCredentials:
        if scopes is None:
            scopes = []
        return _ServiceAccountCredentials(
            service_account_id=credentials["client_id"],
            service_account_email=credentials["client_email"],
            private_key_id=credentials["private_key_id"],
            private_key_pkcs8_text=credentials["private_key"],
            scopes=scopes,
        )


# Silence Google API client verbose spamming
logging.getLogger("googleapiclient.discovery_cache").setLevel(logging.ERROR)
logging.getLogger("googleapiclient").setLevel(logging.WARNING)
logging.getLogger("oauth2client").setLevel(logging.WARNING)

# googleapiclient download performs some 3-4 times better with 50 MB chunk size than 5 MB chunk size;
# but decrypting/decompressing big chunks needs a lot of memory so use smaller chunks on systems with less
# than 2 GB RAM
DOWNLOAD_CHUNK_SIZE = 1024 * 1024 * 5 if (get_total_memory() or 0) < 2048 else 1024 * 1024 * 50
UPLOAD_CHUNK_SIZE = 1024 * 1024 * 5


def get_credentials(
    credential_file: Optional[TextIO] = None, credentials: Optional[dict[str, Any]] = None
) -> GoogleCredentials:
    if credential_file:
        return GoogleCredentials.from_stream(credential_file)

    if credentials and credentials["type"] == "service_account":
        return ServiceAccountCredentials_from_dict(
            credentials,
            scopes=["https://www.googleapis.com/auth/cloud-platform"],
        )

    if credentials and credentials["type"] == "authorized_user":
        return GoogleCredentials(
            access_token=None,
            client_id=credentials["client_id"],
            client_secret=credentials["client_secret"],
            refresh_token=credentials["refresh_token"],
            token_expiry=None,
            token_uri=GOOGLE_TOKEN_URI,
            user_agent="pghoard",
        )

    return GoogleCredentials.get_application_default()


def base64_to_hex(b64val: Union[str, bytes]) -> str:
    if isinstance(b64val, str):
        b64val = b64val.encode("ascii")
    rawval = codecs.decode(b64val, "base64")
    hexval = codecs.encode(rawval, "hex")
    return hexval.decode("ascii")


@dataclasses.dataclass
class Reporter:
    """Used for storing default and also reporting them accordingly

    the whole point of this class is to handle different cases
    - when file size is too small, reporting *_CHUNK_SIZE is wrong as 1000's of
      small files over a long period (may be a month) is a lot of error
    - Same is the case of _retry_on_reset
    - This error extrapolates when replication is being used

    size should ideally be min(real_size, *_CHUNK_SIZE). real_size is avaiable
    when download or upload (from string or file), but when using FD, size is
    not known so assuming something too big for small files is almost wrong and
    in case it is bigger than CHUNKED then status reporting (from the
    googleclient lib) should help get the correct sizes

    """

    operation: StorageOperation
    size: Optional[int] = None
    progress_prev: int = 0

    def report(self, stats: StatsClient) -> None:
        # reports the default.
        # for sized operation, reporting after the operation is fine as _retry_on_reset will
        # have reported the something already
        # and if the operation eventually succeeds then report_status() or this will take care
        # of it.
        SIZED_OPERATIONS = {
            StorageOperation.store_file: UPLOAD_CHUNK_SIZE,
            StorageOperation.get_file: DOWNLOAD_CHUNK_SIZE,
        }

        if self.operation in SIZED_OPERATIONS:
            size = self.size if self.size is not None else SIZED_OPERATIONS[self.operation]
            stats.operation(operation=self.operation, size=size)
        else:
            stats.operation(operation=self.operation)

    def report_status(self, stats: StatsClient, status: Union[MediaUploadProgress, MediaDownloadProgress]) -> None:
        stats.operation(operation=self.operation, size=status.resumable_progress - self.progress_prev)
        self.progress_prev = status.resumable_progress


class Config(StorageModel):
    project_id: str
    bucket_name: str
    credential_file: Optional[str] = None
    credentials: Optional[Dict[str, Any]] = None
    proxy_info: Optional[ProxyInfo] = None
    prefix: Optional[str] = None


ResType = TypeVar("ResType")


class GoogleTransfer(BaseTransfer[Config]):
    config_model = Config

    def __init__(
        self,
        project_id: str,
        bucket_name: str,
        credential_file: Optional[TextIO] = None,
        credentials: Optional[dict[str, Any]] = None,
        prefix: Optional[str] = None,
        proxy_info: Optional[dict[str, Union[str, int]]] = None,
        notifier: Optional[Notifier] = None,
        statsd_info: Optional[StatsdConfig] = None,
    ) -> None:
        super().__init__(prefix=prefix, notifier=notifier, statsd_info=statsd_info)
        self.project_id = project_id
        self.proxy_info = proxy_info
        self.google_creds = get_credentials(credential_file=credential_file, credentials=credentials)
        self.gs: Optional[Resource] = self._init_google_client()
        self.gs_object_client = None
        self.bucket_name = self.get_or_create_bucket(bucket_name)
        self.log.debug("GoogleTransfer initialized")

    def _init_google_client(self) -> Resource:
        start_time = time.monotonic()
        delay = 2
        while True:
            http = build_http()
            if self.proxy_info:
                if self.proxy_info.get("type") == "socks5":
                    proxy_type = httplib2.socks.PROXY_TYPE_SOCKS5  # type: ignore [attr-defined]
                else:
                    proxy_type = httplib2.socks.PROXY_TYPE_HTTP  # type: ignore [attr-defined]

                http.proxy_info = httplib2.ProxyInfo(
                    proxy_type,
                    self.proxy_info["host"],
                    self.proxy_info["port"],
                    proxy_user=self.proxy_info.get("user"),
                    proxy_pass=self.proxy_info.get("pass"),
                )

            http = self.google_creds.authorize(http)

            try:
                # sometimes fails: httplib2.ServerNotFoundError: Unable to find the server at www.googleapis.com
                # https://googleapis.github.io/google-api-python-client/docs/dyn/storage_v1.html
                return build("storage", "v1", http=http)
            except (httplib2.ServerNotFoundError, socket.timeout):
                if time.monotonic() - start_time > 600:
                    raise

            # retry on DNS issues
            time.sleep(delay)
            delay = delay * 2

    @contextmanager
    def _object_client(self, *, not_found: Optional[str] = None) -> Iterator[Any]:
        """(Re-)initialize object client if required, handle 404 errors gracefully and reset the client on
        server errors.  Server errors have been shown to be caused by invalid state in the client and do not
        seem to be resolved without resetting."""
        if self.gs_object_client is None:
            if self.gs is None:
                self.gs = self._init_google_client()
            # https://googleapis.github.io/google-api-python-client/docs/dyn/storage_v1.objects.html
            self.gs_object_client = self.gs.objects()  # type: ignore [attr-defined] # pylint: disable=no-member
        try:
            yield self.gs_object_client
        except HttpError as ex:
            if ex.resp["status"] == "404" and not_found is not None:
                raise FileNotFoundFromStorageError(not_found)
            if ex.resp["status"] >= "500" and ex.resp["status"] <= "599":
                self.log.error("Received server error %r, resetting Google API client", ex.resp["status"])
                self.gs = None
                self.gs_object_client = None
            raise

    def _retry_on_reset(self, request: HttpRequest, action: Callable[[], ResType], retry_reporter: Reporter) -> ResType:
        retries = 60
        retry_wait = 2.0
        while True:
            try:
                return action()
            except (IncompleteRead, HttpError, ssl.SSLEOFError, socket.timeout, OSError, socket.gaierror) as ex:
                # Note that socket.timeout and ssl.SSLEOFError inherit from OSError
                # and the order of handling the errors here needs to be correct
                if not retries:
                    raise
                elif isinstance(ex, (IncompleteRead, socket.timeout, ssl.SSLEOFError, BrokenPipeError)):
                    pass  # just retry with the same sleep amount
                elif isinstance(ex, HttpError):
                    # https://cloud.google.com/storage/docs/json_api/v1/status-codes
                    # https://cloud.google.com/storage/docs/exponential-backoff
                    if ex.resp["status"] not in ("429", "500", "502", "503", "504"):  # pylint: disable=no-member
                        raise
                    retry_wait = min(10.0, max(1.0, retry_wait * 2) + random.random())
                # httplib2 commonly fails with Bad File Descriptor and Connection Reset
                elif isinstance(ex, OSError) and ex.errno not in [errno.EAGAIN, errno.EBADF, errno.ECONNRESET]:
                    raise
                # getaddrinfo sometimes fails with "Name or service not known"
                elif isinstance(ex, socket.gaierror) and ex.errno != socket.EAI_NONAME:
                    raise

                self.log.warning("%s failed: %s (%s), retrying in %.2fs", action, ex.__class__.__name__, ex, retry_wait)

            retry_reporter.report(self.stats)

            # we want to reset the http connection state in case of error
            if request and hasattr(request, "http"):
                request.http.connections.clear()  # reset connection cache

            retries -= 1
            time.sleep(retry_wait)

    def copy_file(
        self, *, source_key: str, destination_key: str, metadata: Optional[Metadata] = None, **_kwargs: Any
    ) -> None:
        source_object = self.format_key_for_backend(source_key)
        destination_object = self.format_key_for_backend(destination_key)
        body = {}
        if metadata is not None:
            body["metadata"] = metadata

        reporter = Reporter(StorageOperation.copy_file)
        with self._object_client(not_found=source_key) as clob:
            # https://googleapis.github.io/google-api-python-client/docs/dyn/storage_v1.objects.html#copy
            request = clob.copy(
                body=body,
                destinationBucket=self.bucket_name,
                destinationObject=destination_object,
                sourceBucket=self.bucket_name,
                sourceObject=source_object,
            )
            result = self._retry_on_reset(request, request.execute, retry_reporter=reporter)
            if result.get("size", None) is not None:
                size = int(result["size"])
                reporter.size = size
                self.notifier.object_copied(key=destination_key, size=size, metadata=metadata)
            reporter.report(self.stats)

    def get_metadata_for_key(self, key: str) -> Metadata:
        path = self.format_key_for_backend(key)
        with self._object_client(not_found=path) as clob:
            return self._metadata_for_key(clob, path)[0]

    def _metadata_for_key(self, clob: Any, key: str) -> tuple[dict[str, Any], int]:
        # https://googleapis.github.io/google-api-python-client/docs/dyn/storage_v1.objects.html#get
        req = clob.get(bucket=self.bucket_name, object=key)

        reporter = Reporter(StorageOperation.get_metadata_for_key)
        obj = self._retry_on_reset(req, req.execute, retry_reporter=reporter)
        reporter.report(self.stats)
        return obj.get("metadata", {}), int(obj["size"])

    def _unpaginate(
        self, domain: Any, initial_op: Callable[[Any], Optional[HttpRequest]], *, on_properties: Iterable[str]
    ) -> Iterator[tuple[str, Any]]:
        """Iterate thru the request pages until all items have been processed"""
        request = initial_op(domain)
        while request is not None:
            reporter = Reporter(StorageOperation.iter_key)
            result = self._retry_on_reset(request, request.execute, retry_reporter=reporter)
            reporter.report(self.stats)
            for on_property in on_properties:
                items = result.get(on_property)
                if items is not None:
                    yield on_property, items
            request = domain.list_next(request, result)

    def iter_key(
        self,
        key: str,
        *,
        with_metadata: bool = True,  # pylint: disable=unused-argument
        deep: bool = False,
        include_key: bool = False,  # pylint: disable=unused-argument, unused-variable
    ) -> Iterator[IterKeyItem]:
        path = self.format_key_for_backend(key, trailing_slash=not include_key)
        self.log.debug("Listing path %r", path)
        with self._object_client() as clob:

            def initial_op(domain: Any) -> HttpRequest:
                if deep:
                    kwargs = {}
                else:
                    kwargs = {"delimiter": "/"}
                # https://googleapis.github.io/google-api-python-client/docs/dyn/storage_v1.objects.html#list
                return domain.list(bucket=self.bucket_name, prefix=path, **kwargs)

            for property_name, items in self._unpaginate(clob, initial_op, on_properties=["items", "prefixes"]):
                if property_name == "items":
                    for item in items:
                        if item["name"].endswith("/"):
                            self.log.warning("list_iter: directory entry %r", item)
                            continue  # skip directory level objects

                        yield IterKeyItem(
                            type=KEY_TYPE_OBJECT,
                            value={
                                "name": self.format_key_from_backend(item["name"]),
                                "size": int(item["size"]),
                                "last_modified": parse_timestamp(item["updated"]),
                                "metadata": item.get("metadata", {}),
                                "md5": base64_to_hex(item["md5Hash"]),
                            },
                        )
                elif property_name == "prefixes":
                    for prefix in items:
                        yield IterKeyItem(type=KEY_TYPE_PREFIX, value=self.format_key_from_backend(prefix).rstrip("/"))
                else:
                    raise NotImplementedError(property_name)

    def delete_key(self, key: str) -> None:
        path = self.format_key_for_backend(key)
        self.log.debug("Deleting key: %r", path)
        with self._object_client(not_found=path) as clob:
            # https://googleapis.github.io/google-api-python-client/docs/dyn/storage_v1.objects.html#delete
            req = clob.delete(bucket=self.bucket_name, object=path)
            reporter = Reporter(StorageOperation.delete_key)
            self._retry_on_reset(req, req.execute, retry_reporter=reporter)
            reporter.report(self.stats)
            self.notifier.object_deleted(key)

    def get_contents_to_fileobj(
        self,
        key: str,
        fileobj_to_store_to: BinaryIO,
        *,
        byte_range: Optional[Tuple[int, int]] = None,
        progress_callback: ProgressProportionCallbackType = None,
    ) -> Metadata:
        path = self.format_key_for_backend(key)
        self.log.debug("Starting to fetch the contents of: %r to %r", path, fileobj_to_store_to)
        next_prog_report = 0.0
        last_log_output = 0.0
        self._validate_byte_range(byte_range)
        with self._object_client(not_found=path) as clob:
            metadata, obj_size = self._metadata_for_key(clob, path)
            if byte_range is None:
                size_to_download = obj_size
            else:
                size_to_download = min(obj_size - byte_range[0], byte_range[1] - byte_range[0] + 1)
            reporter = Reporter(
                StorageOperation.get_file,
                size=min(size_to_download, DOWNLOAD_CHUNK_SIZE),
            )
            # https://googleapis.github.io/google-api-python-client/docs/dyn/storage_v1.objects.html#get_media
            req: HttpRequest = clob.get_media(bucket=self.bucket_name, object=path)
            download: MediaDownloadProtocol
            if byte_range is None:
                download = MediaIoBaseDownload(fileobj_to_store_to, req, chunksize=DOWNLOAD_CHUNK_SIZE)
            else:
                download = MediaIoBaseDownloadWithByteRange(
                    fileobj_to_store_to, req, chunksize=DOWNLOAD_CHUNK_SIZE, byte_range=byte_range
                )

            done = False
            while not done:
                status, done = self._retry_on_reset(req, download.next_chunk, retry_reporter=reporter)
                if status:
                    reporter.report_status(self.stats, status)
                    progress_pct = status.progress() * 100
                    now = time.monotonic()
                    if (now - last_log_output) >= 5.0:
                        self.log.debug("Download of %r: %d%%", path, progress_pct)
                        last_log_output = now

                    if progress_callback and progress_pct > next_prog_report:
                        progress_callback(progress_pct, 100)
                        next_prog_report = progress_pct + 0.1
                elif done:
                    reporter.report_status(self.stats, MediaDownloadProgress(size_to_download, size_to_download))
                else:
                    reporter.report(self.stats)
            if progress_callback:
                progress_callback(100, 100)
            return metadata

    def get_file_size(self, key: str) -> int:
        path = self.format_key_for_backend(key)
        reporter = Reporter(StorageOperation.get_file_size)
        with self._object_client(not_found=path) as clob:
            # https://googleapis.github.io/google-api-python-client/docs/dyn/storage_v1.objects.html#get
            req = clob.get(bucket=self.bucket_name, object=path)
            obj = self._retry_on_reset(req, req.execute, retry_reporter=reporter)
            reporter.report(self.stats)
            return int(obj["size"])

    def _upload(
        self,
        upload: MediaUpload,
        key: str,
        metadata: Metadata,
        extra_props: Optional[dict[str, Any]],
        cache_control: Optional[str],
        reporter: Reporter,
        upload_progress_fn: IncrementalProgressCallbackType = None,
    ) -> dict[str, str]:
        path = self.format_key_for_backend(key)
        self.log.debug("Starting to upload %r", path)
        body: dict[str, Any] = {"metadata": metadata}
        if extra_props:
            body.update(extra_props)
        if cache_control is not None:
            body["cacheControl"] = cache_control

        last_log_output = 0.0
        with self._object_client() as clob:
            # https://googleapis.github.io/google-api-python-client/docs/dyn/storage_v1.objects.html#insert
            req = clob.insert(bucket=self.bucket_name, name=path, media_body=upload, body=body)
            response = None
            while response is None:
                status, response = self._retry_on_reset(req, req.next_chunk, retry_reporter=reporter)
                if status:
                    reporter.report_status(self.stats, status)
                    now = time.monotonic()
                    if (now - last_log_output) >= 5.0:
                        self.log.debug(
                            "Upload of %r to %r: %d%%, %s bytes",
                            upload,
                            path,
                            status.progress() * 100,
                            status.resumable_progress,
                        )
                        last_log_output = now

                    if upload_progress_fn:
                        upload_progress_fn(status.resumable_progress)
                elif response is not None:
                    reporter.report_status(self.stats, MediaUploadProgress(int(response["size"]), int(response["size"])))
                else:
                    reporter.report(self.stats)
        return response

    # pylint: disable=arguments-differ
    def store_file_from_disk(
        self,
        key: str,
        filepath: AnyPath,
        metadata: Optional[Metadata] = None,
        *,
        cache_control: Optional[str] = None,
        mimetype: Optional[str] = None,
        multipart: Optional[bool] = None,
        progress_fn: ProgressProportionCallbackType = None,
        extra_props: Optional[dict[str, Any]] = None,  # pylint: disable=arguments-differ
    ) -> None:  # pylint: disable=unused-argument
        # TODO: extra_props seems to be used only to set cacheControl in pghoard tests.
        #
        # When that is gone (.. long enough ..), we could get rid of
        # this whole function and use the superclass default which
        # falls back to store_file_object.
        size = os.path.getsize(filepath)
        with open(filepath, "rb") as fd:
            metadata = metadata or {}
            metadata.setdefault("Content-Length", size)
            self.store_file_object(
                key,
                fd,
                cache_control=cache_control,
                metadata=metadata,
                mimetype=mimetype,
                multipart=multipart,
                upload_progress_fn=self._incremental_to_proportional_progress(cb=progress_fn, size=size),
                extra_props=extra_props,
            )

    def store_file_object(
        self,
        key: str,
        fd: BinaryIO,
        metadata: Optional[Metadata] = None,
        *,
        cache_control: Optional[str] = None,
        mimetype: Optional[str] = None,
        multipart: Optional[bool] = None,  # pylint: disable=unused-argument
        upload_progress_fn: IncrementalProgressCallbackType = None,
        extra_props: Optional[dict[str, Any]] = None,  # pylint: disable=arguments-differ
    ) -> None:
        mimetype = mimetype or "application/octet-stream"
        sanitized_metadata = self.sanitize_metadata(metadata)
        reporter = Reporter(StorageOperation.store_file)
        result = self._upload(
            MediaStreamUpload(fd, chunk_size=UPLOAD_CHUNK_SIZE, mime_type=mimetype, name=key),
            key,
            sanitized_metadata,
            extra_props,
            cache_control=cache_control,
            upload_progress_fn=upload_progress_fn,
            reporter=reporter,
        )
        self.notifier.object_created(key=key, size=int(result["size"]), metadata=sanitized_metadata)

    def get_or_create_bucket(self, bucket_name: str) -> str:
        """Look up the bucket if it already exists and try to create the
        bucket in case it doesn't.  Note that we can't just always try to
        unconditionally create the bucket as Google imposes a strict rate
        limit on bucket creation operations, even if it doesn't result in a
        new bucket.

        Quietly handle the case where the bucket already exists to avoid
        race conditions.  Note that we'll get a 400 Bad Request response for
        invalid bucket names ("Invalid bucket name") as well as for invalid
        project ("Invalid argument"), try to handle both gracefully."""
        start_time = time.time()
        gs_buckets = self.gs.buckets()  # type: ignore [union-attr] # pylint: disable=no-member
        try:
            request = gs_buckets.get(bucket=bucket_name)
            reporter = Reporter(StorageOperation.head_request)
            self._retry_on_reset(request, request.execute, retry_reporter=reporter)
            reporter.report(self.stats)
            self.log.debug("Bucket: %r already exists, took: %.3fs", bucket_name, time.time() - start_time)
        except HttpError as ex:
            if ex.resp["status"] == "404":
                pass  # we need to create it
            elif ex.resp["status"] == "403":
                raise InvalidConfigurationError("Bucket {0!r} exists but isn't accessible".format(bucket_name))
            else:
                raise
        else:
            return bucket_name

        try:
            req = gs_buckets.insert(project=self.project_id, body={"name": bucket_name})
            reporter = Reporter(StorageOperation.create_bucket)
            self._retry_on_reset(req, req.execute, retry_reporter=reporter)
            reporter.report(self.stats)
            self.log.debug("Created bucket: %r successfully, took: %.3fs", bucket_name, time.time() - start_time)
        except HttpError as ex:
            error = json.loads(ex.content.decode("utf-8"))["error"]
            if error["message"].startswith("You already own this bucket"):
                self.log.debug("Bucket: %r already exists, took: %.3fs", bucket_name, time.time() - start_time)
            elif error["message"] == "Invalid argument.":
                raise InvalidConfigurationError("Invalid project id {0!r}".format(self.project_id))
            elif error["message"].startswith("Invalid bucket name"):
                raise InvalidConfigurationError("Invalid bucket name {0!r}".format(bucket_name))
            else:
                raise

        return bucket_name


class MediaStreamUpload(MediaUpload):
    """Support streaming arbitrary amount of data from non-seekable object supporting read method."""

    def __init__(self, fd: BinaryIO, *, chunk_size: int, mime_type: str, name: str) -> None:
        self._data = b""
        self._next_chunk = b""
        self._chunk_size = chunk_size
        self._fd = fd
        self._mime_type = mime_type
        self._name = name
        self._position: Optional[int] = None

    def chunksize(self) -> int:  # type: ignore [override]
        return self._chunk_size

    def mimetype(self) -> str:
        return self._mime_type

    def size(self) -> Optional[int]:  # type: ignore [override]
        self.peek()
        if len(self._next_chunk) < self.peeksize:
            # The total file size should be returned if we have hit the final chunk.
            return (self._position or 0) + len(self._data) + len(self._next_chunk)
        return None

    def resumable(self) -> bool:
        return True

    @property
    def peeksize(self) -> int:
        # Using 1 extra byte to avoid perfectly aligned file
        return self._chunk_size + 1

    def peek(self) -> None:
        """try to top up some data into _next_chunk"""
        if len(self._next_chunk) < self.peeksize:
            # top-up the _next_chunk
            self._next_chunk = self._read_bytes(self.peeksize - len(self._next_chunk), initial_data=self._next_chunk)

    # second parameter is length but baseclass incorrectly names it end
    def getbytes(self, begin: int, length: int) -> bytes:  # type: ignore [override] # pylint: disable=arguments-differ
        if begin < (self._position or 0):
            msg = "Requested position {} for {!r} precedes already fulfilled position {}".format(
                begin, self._name, self._position
            )
            raise IndexError(msg)
        elif begin > (self._position or 0) + len(self._data):
            msg = "Requested position {} for {!r} has gap from previous position {} and {} byte chunk".format(
                begin, self._name, self._position, len(self._data)
            )
            raise IndexError(msg)

        if self._position is None or begin == self._position + len(self._data):
            if length <= len(self._next_chunk):
                self._data = self._next_chunk[:length]
                self._next_chunk = self._next_chunk[length:]
            else:
                self._data = self._read_bytes(length - len(self._next_chunk), initial_data=self._next_chunk)
                self._next_chunk = b""
        elif begin != self._position or length > len(self._data):
            retain_chunk = self._data[begin - self._position :]
            bytes_remaining = length - len(retain_chunk)
            if 0 < bytes_remaining <= len(self._next_chunk):
                self._data = retain_chunk + self._next_chunk[:bytes_remaining]
                self._next_chunk = self._next_chunk[bytes_remaining:]
            elif bytes_remaining > len(self._next_chunk):
                retain_chunk += self._next_chunk
                self._next_chunk = b""
                self._data = self._read_bytes(length - len(retain_chunk), initial_data=retain_chunk)

        self.peek()
        self._position = begin
        return self._data

    def has_stream(self) -> bool:
        return False

    def stream(self) -> BinaryIO:  # type: ignore [override]
        raise NotImplementedError

    def _read_bytes(self, length: int, *, initial_data: Optional[bytes] = None) -> bytes:
        bytes_remaining = length
        read_results = []
        if initial_data:
            read_results.append(initial_data)
        while bytes_remaining > 0:
            data = self._fd.read(bytes_remaining)
            if data:
                read_results.append(data)
                bytes_remaining -= len(data)
            else:
                break

        if not read_results:
            return b""
        elif len(read_results) == 1:
            return read_results[0]
        else:
            return b"".join(read_results)


class MediaDownloadProtocol(Protocol):
    def next_chunk(self) -> tuple[MediaDownloadProgress, bool]:
        ...


class MediaIoBaseDownloadWithByteRange:
    """This class is mostly a copy of the googleapiclient's MediaIOBaseDownload class,
    but with the addition of the support for fetching a specific byte_range.

    """

    def __init__(
        self,
        fd: BinaryIO,
        request: HttpRequest,
        chunksize: int = DOWNLOAD_CHUNK_SIZE,
        *,
        byte_range: tuple[int, int],
    ) -> None:
        """Constructor.

        Args:
          fd: io.Base or file object, The stream in which to write the downloaded
            bytes.
          request: googleapiclient.http.HttpRequest, the media request to perform in
            chunks.
          chunksize: int, File will be downloaded in chunks of this many bytes.
          byte_range: tuple[int, int], The byterange to fetch
        """
        self._fd = fd
        self._http = request.http
        self._uri = request.uri
        self._chunksize = chunksize
        self._start_position, self._end_position = byte_range
        self._num_bytes_downloaded = 0
        self._range_size = self._end_position - self._start_position + 1
        if self._range_size < 0:
            raise InvalidByteRangeError(f"Invalid byte_range: {byte_range}. Start must be < end.")
        self._done = False

        self._headers = {}
        for k, v in request.headers.items():
            # allow users to supply custom headers by setting them on the request
            # but strip out the ones that are set by default on requests generated by
            # API methods like Drive's files().get(fileId=...)
            if k.lower() not in ("accept", "accept-encoding", "user-agent"):
                self._headers[k] = v

    def next_chunk(self) -> tuple[MediaDownloadProgress, bool]:
        """Get the next chunk of the download.

        Returns:
          (status, done): The value of done will be True when the media has been fully
             downloaded or the total size of the media is unknown.

        Raises:
          googleapiclient.errors.HttpError if the response was not a 2xx (or a 416 is received and the file is empty)
          httplib2.HttpLib2Error if a transport error has occurred.
        """
        headers = self._headers.copy()
        chunk_start = self._num_bytes_downloaded + self._start_position
        chunk_end = chunk_start + self._chunksize - 1
        if self._end_position < chunk_end:
            chunk_end = self._end_position
        headers["range"] = f"bytes={chunk_start}-{chunk_end}"
        resp, content = self._http.request(self._uri, "GET", headers=headers)

        total_size: Optional[int] = None
        if resp.status in (200, 206):
            if "content-location" in resp and resp["content-location"] != self._uri:
                self._uri = resp["content-location"]
            self._num_bytes_downloaded += len(content)
            self._fd.write(content)

            if "content-range" in resp:
                total_size = get_total_size_from_content_range(resp["content-range"])
            elif "content-length" in resp:
                # By RFC 9110 if we end up here this is a 200 OK response and this is the total size of the object
                total_size = int(resp["content-length"])

            size_to_download = (
                self._range_size if total_size is None else min(self._range_size, total_size - self._start_position)
            )
            if self._num_bytes_downloaded == size_to_download:
                self._done = True
            return MediaDownloadProgress(self._num_bytes_downloaded, size_to_download), self._done
        elif resp.status == 416:
            # 416 is Range Not Satisfiable
            # This typically occurs with a zero byte file
            total_size = get_total_size_from_content_range(resp["content-range"])
            if total_size == 0:
                self._done = True
                return MediaDownloadProgress(self._num_bytes_downloaded, total_size), self._done
        raise HttpError(resp, content, uri=self._uri)



"""
Copyright (c) 2020 Aiven Ltd
See LICENSE for details

StatsD client

Supports telegraf's statsd protocol extension for 'key=value' tags:

  https://github.com/influxdata/telegraf/tree/master/plugins/inputs/statsd

This is combination of:
- pghoard base (pghoard.metrics.statsd)
- myhoard timing_manager method
- pydantic configuration + explicit typing

"""

from __future__ import annotations

from contextlib import asynccontextmanager, contextmanager
from enum import Enum
from typing import AsyncIterator, Dict, Iterator, Optional, Union

import pydantic
import socket
import time


class MessageFormat(str, Enum):
    datadog = "datadog"
    telegraf = "telegraf"

    def __str__(self) -> str:
        return str(self.value)


Tags = Dict[str, Union[int, str, None]]


class StatsdConfig(pydantic.BaseModel):
    host: str = "127.0.0.1"
    port: int = 8125
    message_format: MessageFormat = MessageFormat.telegraf
    tags: Tags = {}
    operation_map: Dict[str, str] = {}

    class Config:
        use_enum_values = True
        extra = "forbid"
        validate_all = True


class StatsClient:
    _enabled = True

    def __init__(self, config: Optional[StatsdConfig]):
        self._operation_map = {}
        if not config:
            self._enabled = False
            return
        if not isinstance(config, StatsdConfig):
            config = StatsdConfig.parse_obj(config)
        self._dest_addr = (config.host, config.port)
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._tags = config.tags
        self._message_format = config.message_format
        self._operation_map = config.operation_map

    @asynccontextmanager
    async def async_timing_manager(self, metric: str, tags: Optional[Tags] = None) -> AsyncIterator[None]:
        with self.timing_manager(metric, tags=tags):
            yield

    @contextmanager
    def timing_manager(self, metric: str, tags: Optional[Tags] = None) -> Iterator[None]:
        start_time = time.monotonic()
        tags = (tags or {}).copy()
        try:
            yield
        except:  # noqa pylint: disable=broad-except,bare-except
            tags["success"] = "0"
            self.timing(metric, time.monotonic() - start_time, tags=tags)
            raise
        tags["success"] = "1"
        self.timing(metric, time.monotonic() - start_time, tags=tags)

    def gauge(self, metric: str, value: Union[int, float], *, tags: Optional[Tags] = None) -> None:
        self._send(metric, b"g", value, tags)

    def increase(self, metric: str, *, inc_value: int = 1, tags: Optional[Tags] = None) -> None:
        self._send(metric, b"c", inc_value, tags)

    def timing(self, metric: str, value: Union[int, float], *, tags: Optional[Tags] = None) -> None:
        self._send(metric, b"ms", value, tags)

    def unexpected_exception(self, ex: BaseException, where: str, *, tags: Optional[Tags] = None) -> None:
        all_tags: Tags = {
            "exception": ex.__class__.__name__,
            "where": where,
        }
        all_tags.update(tags or {})
        self.increase("exception", tags=all_tags)

    def operation(self, operation: str, *, count: int = 1, size: Union[int, None] = None) -> None:
        tags: Tags = {"operation": self._operation_map.get(str(operation), str(operation))}
        self.increase("rohmu_operation_count", tags=tags, inc_value=count)
        if size is not None:
            self.increase("rohmu_operation_size", tags=tags, inc_value=size)

    def _send(self, metric: str, metric_type: bytes, value: Union[int, float], tags: Optional[Tags]) -> None:
        if not self._enabled:
            # stats sending is disabled
            return

        # telegraf format: "user.logins,service=payroll,region=us-west:1|c"
        # datadog format: metric.name:value|type|@sample_rate|#tag1:value,tag2
        #                 http://docs.datadoghq.com/guides/dogstatsd/#datagram-format

        parts = [metric.encode("utf-8"), b":", str(value).encode("utf-8"), b"|", metric_type]
        send_tags = self._tags.copy()
        send_tags.update(tags or {})
        if self._message_format == MessageFormat.datadog:
            for index, (tag, val) in enumerate(send_tags.items()):
                if index == 0:
                    separator = "|#"
                else:
                    separator = ","
                if val is None:
                    pattern = "{}{}"
                else:
                    pattern = "{}{}:{}"
                parts.append(pattern.format(separator, tag, val).encode("utf-8"))
        elif self._message_format == MessageFormat.telegraf:
            for tag, val in send_tags.items():
                parts.insert(1, f",{tag}={val}".encode("utf-8"))
        else:
            raise NotImplementedError("Unsupported message format")

        self._socket.sendto(b"".join(parts), self._dest_addr)



# Copyright (c) 2023 Aiven, Helsinki, Finland. https://aiven.io/


from rohmu.common.statsd import StatsdConfig
from rohmu.notifier.interface import Notifier
from typing import Optional

import enum
import pydantic


class StorageOperation(str, enum.Enum):
    iter_key = "iter_key"
    copy_file = "copy_file"
    delete_key = "delete_key"
    get_file = "get_file"
    get_file_size = "get_file_size"
    get_metadata_for_key = "get_metadata_for_key"
    list_path = "list_path"
    list_iter = "list_iter"
    store_file = "store_file"
    metadata_for_key = "metadata_for_key"
    head_request = "head_request"
    create_bucket = "create_bucket"

    # These are S3-only but their use (and especially failures) are interesting
    create_multipart_upload = "create_multipart_upload"
    multipart_aborted = "multipart_aborted"
    multipart_complete = "multipart_complete"

    def __str__(self) -> str:
        return str(self.value)


class ProxyType(str, enum.Enum):
    socks5 = "socks5"
    http = "http"


class RohmuModel(pydantic.BaseModel):
    class Config:
        # As we're keen to both export and decode json, just using
        # enum values for encode/decode is much saner than the default
        # enumname.value (it is also slightly less safe but oh well)
        use_enum_values = True

        # Extra values should be errors, as they are most likely typos
        # which lead to grief when not detected. However, if we ever
        # start deprecating some old fields and not wanting to parse
        # them, this might need to be revisited.
        extra = "forbid"

        # Validate field default values too
        validate_all = True

        # Validate also assignments
        validate_assignment = True


class ProxyInfo(RohmuModel):
    host: str
    port: int
    type: ProxyType
    user: Optional[str]
    password: Optional[str] = pydantic.Field(None, alias="pass")


class StorageModel(pydantic.BaseModel):
    notifier: Optional[Notifier] = None
    statsd_info: Optional[StatsdConfig] = None

    class Config:
        use_enum_values = True
        # Notifier Type does not have validation (yet)
        arbitrary_types_allowed = True
        # extra fields would not be accepted by corresponding __init__ methods anyways
        extra_forbid = True



"""
rohmu - exception classes

Copyright (c) 2016 Ohmu Ltd
See LICENSE for details
"""


class Error(Exception):
    """Generic exception"""


class StorageError(Error):
    """Storage exception"""


class FileNotFoundFromStorageError(StorageError):
    """File not found from remote storage"""


class InvalidConfigurationError(Error):
    """Invalid configuration"""


class LocalFileIsRemoteFileError(StorageError):
    """File transfer operation source and destination point to the same file"""


class MissingLibraryError(Exception):
    """Missing dependency library"""


class MaybeRecoverableError(Error):
    """An error that may be recoverable"""


class UninitializedError(Error):
    """Error trying to access an uninitialized resource"""


class InvalidByteRangeError(Error):
    """Error specifying a content-range in a request"""


class InvalidTransferError(Error):
    """You tried to access a transfer object that you already returned to the pool"""



from __future__ import annotations

from types import TracebackType
from typing import Any, Dict, Optional, Protocol, Type, TYPE_CHECKING, Union
from typing_extensions import Buffer

if TYPE_CHECKING:
    from array import array
    from os import PathLike

    import ctypes

Metadata = Dict[str, Any]

AnyPath = Union[str, bytes, "PathLike[str]", "PathLike[bytes]"]

BinaryData = Buffer

StrOrPathLike = Union[str, "PathLike[str]"]


class HasFileno(Protocol):
    def fileno(self) -> int:
        ...


class HasRead(Protocol):
    def read(self, n: Optional[int] = -1) -> bytes:
        ...


class HasWrite(Protocol):
    def write(self, data: BinaryData) -> int:
        ...


class FileLike(Protocol):
    def __enter__(self) -> FileLike:
        ...

    def __exit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> None:
        ...

    def read(self, n: Optional[int] = -1) -> bytes:
        ...

    def flush(self) -> None:
        ...

    def write(self, data: BinaryData) -> int:
        ...

    def close(self) -> None:
        ...

    def fileno(self) -> int:
        ...

    def tell(self) -> int:
        ...

    def seek(self, offset: int, whence: int) -> int:
        ...


class Compressor(Protocol):
    def compress(self, data: bytes) -> bytes:
        ...

    def flush(self) -> bytes:
        ...


class Decompressor(Protocol):
    def decompress(self, data: bytes) -> bytes:
        ...



"""
rohmu - common utility functions

Copyright (c) 2022 Ohmu Ltd
See LICENSE for details
"""
from itertools import islice
from rohmu.typing import HasFileno
from typing import Generator, Iterable, Optional, Tuple, TypeVar, Union

import fcntl
import logging
import os
import platform

LOG = logging.getLogger("rohmu.util")


def increase_pipe_capacity(*pipes: Union[int, HasFileno]) -> None:
    if platform.system() != "Linux":
        return
    try:
        with open("/proc/sys/fs/pipe-max-size", "r") as f:
            pipe_max_size = int(f.read())
    except FileNotFoundError:
        return
    # Attempt to get as big pipe as possible; as Linux pipe usage quotas are
    # account wide (and not visible to us), brute-force attempting is
    # the best we can do.
    #
    # F_SETPIPE_SZ can also return EBUSY if trying to shrink pipe from
    # what is in the buffer (not true in our case as pipe should be
    # growing), or ENOMEM, and we bail in both of those cases.
    for pipe in pipes:
        for shift in range(0, 16):
            size = pipe_max_size >> shift
            if size <= 65536:
                # Default size
                LOG.warning("Unable to grow pipe buffer at all, performance may suffer")
                return
            try:
                fcntl.fcntl(pipe, 1031, pipe_max_size)  # F_SETPIPE_SZ
                break
            except PermissionError:
                pass


def set_stream_nonblocking(stream: HasFileno) -> None:
    fd = stream.fileno()
    fl = fcntl.fcntl(fd, fcntl.F_GETFL)
    fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)


T = TypeVar("T")


def batched(iterable: Iterable[T], n: int) -> Generator[Tuple[T, ...], None, None]:
    "Batch data into tuples of length n. The last batch may be shorter."
    # batched('ABCDEFG', 3) --> ABC DEF G
    # NOTE: can replace with itertools version once on python 3.12
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    batch = tuple(islice(it, n))
    while batch:
        yield batch
        batch = tuple(islice(it, n))


def get_total_size_from_content_range(content_range: str) -> Optional[int]:
    length = content_range.rsplit("/", 1)[1]
    # RFC 9110 section 14.4 specifies that the * can be returned when the total length is unknown
    return int(length) if length != "*" else None



from contextlib import contextmanager
from pathlib import Path
from typing import BinaryIO, Callable, cast, ContextManager, Generator, Optional, overload, TextIO, Union
from typing_extensions import TypeAlias

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore [assignment]

import errno
import os

Write: TypeAlias = Literal["w"]
WriteBinary: TypeAlias = Literal["wb"]


def _fd_close_quietly(fd: int) -> None:
    try:
        os.close(fd)
    except OSError as e:
        if e.errno == errno.EBADF:
            # closed already
            return
        raise


@overload
def atomic_opener(
    final_path: Path,
    mode: WriteBinary,
    encoding: Optional[str] = None,
    _fd_spy: Callable[[int], None] = lambda unused: None,
    _after_link_hook: Callable[[], None] = lambda: None,
) -> ContextManager[BinaryIO]:
    ...


@overload
def atomic_opener(
    final_path: Path,
    mode: Write,
    encoding: Optional[str] = None,
    _fd_spy: Callable[[int], None] = lambda unused: None,
    _after_link_hook: Callable[[], None] = lambda: None,
) -> ContextManager[TextIO]:
    ...


def atomic_opener(
    final_path: Path,
    mode: str,
    encoding: Optional[str] = None,
    _fd_spy: Callable[[int], None] = lambda unused: None,
    _after_link_hook: Callable[[], None] = lambda: None,
) -> ContextManager[Union[TextIO, BinaryIO]]:
    return _atomic_opener(final_path, mode, encoding, _fd_spy, _after_link_hook)


@contextmanager
def _atomic_opener(
    final_path: Path,
    mode: str,
    encoding: Optional[str] = None,
    _fd_spy: Callable[[int], None] = lambda unused: None,
    _after_link_hook: Callable[[], None] = lambda: None,
) -> Generator[Union[TextIO, BinaryIO], None, None]:
    """
    Creates a file object for writing which will only appear on the filesystem if the context manager succeeds.

    It's designed to prevent partial writes (file will either be available with full content, or will be unavailable)
    and the need for manual cleanup.
    """
    parent_dir = final_path.parent
    if not parent_dir.exists():
        raise IOError(f"Parent directory '{parent_dir}' must exist but is missing")
    if "w" not in mode:
        raise ValueError("Write mode must be used to make actual sense")

    fd = os.open(str(parent_dir), os.O_TMPFILE | os.O_RDWR, 0o600)
    _fd_spy(fd)
    try:
        file_obj = os.fdopen(fd, mode=mode, encoding=encoding)
    except Exception:  # pylint: disable=broad-except
        # when passing wrong mode, os.fdopen won't close input fd. When passing wrong encoding, it will.
        _fd_close_quietly(fd)
        raise

    try:
        if "b" in mode:
            yield cast(BinaryIO, file_obj)
        else:
            yield cast(TextIO, file_obj)
        file_obj.flush()
        path = "/proc/self/fd/{0}".format(fd)
        os.link(path, str(final_path), src_dir_fd=0, follow_symlinks=True)
        _after_link_hook()
    finally:
        file_obj.close()



"""
rohmu - object_storage.base

Copyright (c) 2016 Ohmu Ltd
Copyright (c) 2022 Aiven, Helsinki, Finland. https://aiven.io/
See LICENSE for details
"""
from __future__ import annotations

from ..common.models import StorageModel
from ..common.statsd import StatsClient, StatsdConfig
from ..errors import FileNotFoundFromStorageError, InvalidByteRangeError, StorageError
from ..notifier.interface import Notifier
from ..notifier.null import NullNotifier
from ..typing import AnyPath, Metadata
from contextlib import suppress
from io import BytesIO
from typing import Any, BinaryIO, Callable, Collection, Generic, Iterator, NamedTuple, Optional, Tuple, Type, TypeVar, Union

import logging
import os
import platform

KEY_TYPE_OBJECT = "object"
KEY_TYPE_PREFIX = "prefix"


class IterKeyItem(NamedTuple):
    type: str
    value: Union[str, dict[str, Any]]


# Percent complete is the ratio of the first argument to the second
ProgressProportionCallbackType = Optional[Callable[[int, int], None]]

# Argument is the additional number of bytes transferred
IncrementalProgressCallbackType = Optional[Callable[[int], None]]


class Config(StorageModel):
    prefix: str
    notifier: Optional[Notifier] = None

    class Config:
        arbitrary_types_allowed = True


StorageModelT = TypeVar("StorageModelT", bound=StorageModel)


class BaseTransfer(Generic[StorageModelT]):
    config_model: Type[StorageModelT]

    is_thread_safe: bool = False

    def __init__(
        self, prefix: Optional[str], notifier: Optional[Notifier] = None, statsd_info: Optional[StatsdConfig] = None
    ) -> None:
        self.log = logging.getLogger(self.__class__.__name__)
        if not prefix:
            prefix = ""
        elif prefix[-1] != "/":
            prefix += "/"
        self.prefix = prefix
        self.notifier = notifier or NullNotifier()
        self.stats = StatsClient(statsd_info)

    @staticmethod
    def _incremental_to_proportional_progress(
        *, size: int, cb: ProgressProportionCallbackType
    ) -> IncrementalProgressCallbackType:
        if cb is None:
            return None

        progress_so_far: int = 0

        def wrapper(progress: int) -> None:
            nonlocal progress_so_far
            progress_so_far += progress
            if cb is not None:
                cb(progress_so_far, size)

        return wrapper

    @staticmethod
    def _proportional_to_incremental_progress(cb: IncrementalProgressCallbackType) -> ProgressProportionCallbackType:
        if cb is None:
            return cb
        last_progress: int = 0

        def wrapper(progress: int, _: int) -> None:
            nonlocal last_progress
            if progress > last_progress:
                if cb is not None:
                    cb(progress - last_progress)
                last_progress = progress

        return wrapper

    @staticmethod
    def _should_multipart(
        *, metadata: Optional[Metadata], chunk_size: int, multipart: Union[bool, None] = None, default: bool
    ) -> bool:
        if multipart is not None:
            return multipart

        # multipart = None; up to us
        size = (metadata or {}).get("Content-Length")
        if size is None:
            # We could actually sniff from fd if it is seekable; left TODO for now.
            return default

        return int(size) > chunk_size

    @classmethod
    def from_model(cls, model: StorageModelT) -> BaseTransfer[StorageModelT]:
        return cls(**model.dict(by_alias=True))

    def copy_file(
        self, *, source_key: str, destination_key: str, metadata: Optional[Metadata] = None, **_kwargs: Any
    ) -> None:
        """Performs remote copy from source key name to destination key name. Key must identify a file, trees
        cannot be copied with this method. If no metadata is given copies the existing metadata."""
        raise NotImplementedError

    def format_key_for_backend(self, key: str, remove_slash_prefix: bool = False, trailing_slash: bool = False) -> str:
        """Add a possible prefix to the key before sending it to the backend"""
        path = self.prefix + key
        if trailing_slash:
            if not path or path[-1] != "/":
                path += "/"
        else:
            path = path.rstrip("/")
        if remove_slash_prefix:  # Azure defines slashes in the beginning as "dirs" for listing purposes
            path = path.lstrip("/")
        return path

    def format_key_from_backend(self, key: str) -> str:
        """Strip the configured prefix from a key retrieved from the backend
        before passing it on to other pghoard code and presenting it to the
        user."""
        if not self.prefix:
            return key
        if not key.startswith(self.prefix):
            raise StorageError("Key {!r} does not start with expected prefix {!r}".format(key, self.prefix))
        return key[len(self.prefix) :]

    def delete_key(self, key: str) -> None:
        raise NotImplementedError

    def delete_keys(self, keys: Collection[str]) -> None:
        """Delete specified keys"""
        for key in keys:
            self.delete_key(key)

    def delete_tree(self, key: str) -> None:
        """Delete all keys under given root key. Basic implementation works by just listing all available
        keys and deleting them individually but storage providers can implement more efficient logic."""
        self.log.debug("Deleting tree: %r", key)
        names = [item["name"] for item in self.list_path(key, with_metadata=False, deep=True)]
        self.delete_keys(names)

    def get_contents_to_file(
        self, key: str, filepath_to_store_to: AnyPath, *, progress_callback: ProgressProportionCallbackType = None
    ) -> Metadata:
        """Write key contents to file pointed by `path` and return metadata.  If `progress_callback` is
        provided it must be a function which accepts two numeric arguments: current state of progress and the
        expected maximum value.  The actual values and value ranges differ per storage provider, some (S3)
        reporting the number of bytes transmitted as the first argument and the total number of expected bytes
        as the second argument, while others (Google) report the progress in percentage as the first value and
        100 as the second value."""
        try:
            with open(filepath_to_store_to, "wb") as fd:
                metadata = self.get_contents_to_fileobj(key, fd, progress_callback=progress_callback)
                return metadata
        except FileNotFoundError as ex:
            raise FileNotFoundFromStorageError from ex
        except:
            with suppress(FileNotFoundError):
                os.unlink(filepath_to_store_to)
            raise

    def get_contents_to_fileobj(
        self,
        key: str,
        fileobj_to_store_to: BinaryIO,
        *,
        byte_range: Optional[Tuple[int, int]] = None,
        progress_callback: ProgressProportionCallbackType = None,
    ) -> Metadata:
        """Like `get_contents_to_file()` but writes to an open file-like object."""
        raise NotImplementedError

    def _validate_byte_range(self, byte_range: Optional[Tuple[int, int]]) -> None:
        if byte_range is not None and byte_range[0] > byte_range[1]:
            raise InvalidByteRangeError(f"Invalid byte_range: {byte_range}. Start must be <= end.")

    def get_contents_to_string(self, key: str, *, byte_range: Optional[Tuple[int, int]] = None) -> tuple[bytes, Metadata]:
        """Returns a tuple (content-byte-string, metadata).

        byte_range can be used to limit the content requested (as per RFC9110 section 14.1.2):
        it defines range of bytes to fetch (inclusive), so e.g. [0, 499] means first 500 bytes.
        If it is not supported for specific storage backend, NotImplementedError will be raised.
        """
        with BytesIO() as buf:
            metadata = self.get_contents_to_fileobj(key, buf, byte_range=byte_range)
            return buf.getvalue(), metadata

    def get_file_size(self, key: str) -> int:
        """Returns an int indicating the size of the file in bytes"""
        # This method isn't currently used by PGHoard itself, it is merely provided
        # for applications that use PGHoard's object storage abstraction layer.
        raise NotImplementedError

    def get_metadata_for_key(self, key: str) -> Metadata:
        raise NotImplementedError

    def list_path(self, key: str, *, with_metadata: bool = True, deep: bool = False) -> list[dict[str, Any]]:
        return list(self.list_iter(key, with_metadata=with_metadata, deep=deep))

    def list_iter(self, key: str, *, with_metadata: bool = True, deep: bool = False) -> Iterator[dict[str, Any]]:
        for item in self.iter_key(key, with_metadata=with_metadata, deep=deep):
            if item.type == KEY_TYPE_OBJECT:
                assert isinstance(item.value, dict)
                yield item.value

    def list_prefixes(self, key: str) -> list[str]:
        return list(self.iter_prefixes(key))

    def iter_prefixes(self, key: str) -> Iterator[str]:
        for item in self.iter_key(key, with_metadata=False):
            if item.type == KEY_TYPE_PREFIX:
                assert isinstance(item.value, str)
                yield item.value

    def iter_key(
        self, key: str, *, with_metadata: bool = True, deep: bool = False, include_key: bool = False
    ) -> Iterator[IterKeyItem]:
        raise NotImplementedError

    def sanitize_metadata(self, metadata: Optional[Metadata], replace_hyphen_with: str = "-") -> dict[str, str]:
        """Convert non-string metadata values to strings and drop null values"""
        return {str(k).replace("-", replace_hyphen_with): str(v) for k, v in (metadata or {}).items() if v is not None}

    def store_file_from_memory(
        self,
        key: str,
        memstring: bytes,
        metadata: Optional[Metadata] = None,
        *,
        cache_control: Optional[str] = None,
        mimetype: Optional[str] = None,
        multipart: Optional[bool] = None,
        progress_fn: ProgressProportionCallbackType = None,
    ) -> None:
        with BytesIO(memstring) as buf:
            size = len(memstring)
            if metadata is None:
                metadata = {"Content-Length": size}
            elif metadata.get("Content-Length") is None:
                metadata = metadata.copy()
                metadata["Content-Length"] = size
            self.store_file_object(
                key,
                buf,
                cache_control=cache_control,
                metadata=metadata,
                mimetype=mimetype,
                multipart=multipart,
                upload_progress_fn=self._incremental_to_proportional_progress(cb=progress_fn, size=size),
            )

    def store_file_from_disk(
        self,
        key: str,
        filepath: AnyPath,
        metadata: Optional[Metadata] = None,
        *,
        cache_control: Optional[str] = None,
        mimetype: Optional[str] = None,
        multipart: Optional[bool] = None,
        progress_fn: ProgressProportionCallbackType = None,
    ) -> None:
        size = os.path.getsize(filepath)
        with open(filepath, "rb") as fd:
            if metadata is None:
                metadata = {"Content-Length": size}
            elif metadata.get("Content-Length") is None:
                metadata = metadata.copy()
                metadata["Content-Length"] = size
            self.store_file_object(
                key,
                fd,
                cache_control=cache_control,
                metadata=metadata,
                mimetype=mimetype,
                multipart=multipart,
                upload_progress_fn=self._incremental_to_proportional_progress(cb=progress_fn, size=size),
            )

    def store_file_object(
        self,
        key: str,
        fd: BinaryIO,
        metadata: Optional[Metadata] = None,
        *,
        cache_control: Optional[str] = None,
        mimetype: Optional[str] = None,
        multipart: Optional[bool] = None,
        upload_progress_fn: IncrementalProgressCallbackType = None,
    ) -> None:
        raise NotImplementedError


def get_total_memory() -> Optional[int]:
    """return total system memory in mebibytes (or None if parsing meminfo fails)"""
    if platform.system() != "Linux":
        return None

    with open("/proc/meminfo", "r") as in_file:
        for line in in_file:
            info = line.split()
            if info[0] == "MemTotal:" and info[-1] == "kB":
                memory_mb = int(int(info[1]) / 1024)
                return memory_mb

    return None



"""
rohmu - transfer_pool

Copyright (c) 2023 Aiven, Helsinki, Finland. https://aiven.io/
See LICENSE for details
"""

from __future__ import annotations

from . import get_transfer as rohmu_get_transfer
from .errors import InvalidTransferError
from .object_storage.base import BaseTransfer, StorageModel
from contextlib import contextmanager
from typing import Any, Callable, Generator, Optional

import heapq
import json
import logging
import threading
import time

LOG = logging.getLogger(__name__)

TRANSFER_MAX_AGE = 60 * 60 * 3
TRANSFER_CACHE_MAX_AGE = TRANSFER_MAX_AGE * 2


class TransferCacheItem:
    create_time: float
    cache_key: str
    transfer: BaseTransfer[StorageModel]
    max_age: float

    def __init__(self, pool_key: str, transfer: BaseTransfer[StorageModel], max_age: float = TRANSFER_MAX_AGE) -> None:
        self.create_time = time.monotonic()
        self.cache_key = pool_key
        self.transfer = transfer
        self.max_age = max_age

    def age(self) -> float:
        return time.monotonic() - self.create_time

    def is_expired(self) -> bool:
        return self.age() > self.max_age

    def __gt__(self, other: TransferCacheItem) -> bool:
        """Ordering by creation time, needed for storing these in a heap."""
        return self.create_time > other.create_time


class _TransferCache:
    transfers_heap: list[TransferCacheItem]
    last_used: float
    max_age: float

    def __init__(self, max_age: float = TRANSFER_CACHE_MAX_AGE) -> None:
        self.transfers_heap = []
        self.last_used = time.monotonic()
        self.max_age = max_age

    def prune_expired(self) -> None:
        while self.transfers_heap and self.transfers_heap[0].is_expired():
            heapq.heappop(self.transfers_heap)

    def get(self) -> Optional[TransferCacheItem]:
        self.last_used = time.monotonic()
        self.prune_expired()
        if self.transfers_heap:
            try:
                return heapq.heappop(self.transfers_heap)
            except IndexError:
                pass
        return None

    def put(self, transfer: TransferCacheItem) -> None:
        self.last_used = time.monotonic()
        if not transfer.is_expired():
            heapq.heappush(self.transfers_heap, transfer)

    def age(self) -> float:
        return time.monotonic() - self.last_used

    def is_expired(self) -> bool:
        return self.age() > self.max_age


class _TransferCacheForThreadSafeTransfer(_TransferCache):
    """This _TransferCache subclass is a simple implementation that works when the transfer is thread-safe.

    In that case we can have a single instance and always return that one.

    No expiration logic is applied to the transfer instance.
    """

    def __init__(self, transfer_item: TransferCacheItem, max_age: float = TRANSFER_CACHE_MAX_AGE) -> None:
        super().__init__(max_age)
        self.transfers_heap.append(transfer_item)

    def get(self) -> Optional[TransferCacheItem]:
        self.last_used = time.monotonic()
        # there will always be only one transfer instance at most in this cache type
        return self.transfers_heap[0]

    def put(self, transfer: TransferCacheItem) -> None:
        self.last_used = time.monotonic()


_BASE_TRANSFER_INSTANCE_ATTRS = {"config_model", "log", "notifier", "prefix", "stats"}
_BASE_TRANSFER_ATTRS = {attr for attr in vars(BaseTransfer) if not attr.startswith("__")} | _BASE_TRANSFER_INSTANCE_ATTRS

# pylint: disable=abstract-method,super-init-not-called
class SafeTransfer(BaseTransfer[StorageModel]):
    """Helper class that helps the users in finding bugs in their code handling transfers.

    This class prevents any call to a transfer instance that was returned to the pool.
    It also logs a warning if the instance is garbage collected (returning it to the pool).

    """

    def __init__(self, item: TransferCacheItem, pool_reclaim: Callable[[TransferCacheItem], None]) -> None:
        self._item = item
        self._pool_reclaim = pool_reclaim
        self._done = False

    def __del__(self) -> None:
        if not self._done:
            LOG.warning("Transfer not marked as done was garbage collected. Returning it to the pool now.")

    def __getattribute__(self, attr: str) -> Any:
        if attr in _BASE_TRANSFER_ATTRS:
            if self._done:
                raise InvalidTransferError("Trying to access transfer instance already returned to the pool")
            else:
                return getattr(self._item.transfer, attr)
        return super().__getattribute__(attr)

    @classmethod
    def from_model(cls, model: StorageModel) -> BaseTransfer[StorageModel]:
        raise InvalidTransferError("You should not call class methods on SafeTransfer instances")

    def return_to_pool(self) -> None:
        """Return the underlying transfer instance to the pool.

        After this call all uses of this transfer instance will raise InvalidTransferError.

        This method can be called multiple times.
        """
        if self._done:
            return
        self._pool_reclaim(self._item)
        self._done = True
        del self._item


class TransferPool:
    def __init__(self, *, max_pool_age: float = TRANSFER_CACHE_MAX_AGE, max_transfer_age: float = TRANSFER_MAX_AGE):
        self._max_pool_age = max_pool_age
        self._max_transfer_age = max_transfer_age
        self._caches: dict[str, _TransferCache] = {}
        self._is_thread_safe: dict[str, bool] = {}
        self._mutex = threading.Lock()

    @contextmanager
    def with_transfer(self, storage_config: dict[str, Any]) -> Generator[BaseTransfer[StorageModel], None, None]:
        """Yield a transfer object according to the provided storage_config.

        The transfer object may be created or cached. While the context manager is running the transfer object
        won't be available to other threads.

        Users of this contextmanager should NOT use the transfer object outside the `with` block
        """
        transfer = self._get(storage_config)
        try:
            yield transfer.transfer
        finally:
            self._put(transfer)

    def get_transfer(self, storage_config: dict[str, Any]) -> SafeTransfer:
        """Returns a transfer object.

        The object returned is a proxy that will forward all accesses to the underlying transfer instance.
        You MUST call `return_to_pool()` method once you are done with this transfer, and any access to the proxy
        afterwards will raise an `InvalidTransferError`.

        """
        return SafeTransfer(item=self._get(storage_config), pool_reclaim=self._put)

    def _prune_expired_caches(self) -> None:
        self._caches = {key: pool for key, pool in self._caches.items() if not pool.is_expired()}

    def _get(self, storage_config: dict[str, Any]) -> TransferCacheItem:
        cache_key = _cache_key(storage_config)
        with self._mutex:
            if (cache := self._caches.get(cache_key)) and (cached_transfer := cache.get()):
                return cached_transfer
            transfer = rohmu_get_transfer(storage_config)
            item = TransferCacheItem(cache_key, transfer, max_age=self._max_transfer_age)
            if transfer.is_thread_safe:
                self._caches[cache_key] = _TransferCacheForThreadSafeTransfer(item, max_age=self._max_pool_age)
            else:
                self._caches[cache_key] = _TransferCache(max_age=self._max_pool_age)

            return item

    def _put(self, transfer: TransferCacheItem) -> None:
        with self._mutex:
            transfer_pool = self._caches.get(transfer.cache_key)
            if transfer_pool:
                transfer_pool.put(transfer)
            self._prune_expired_caches()


def _cache_key(storage_config: dict[str, Any]) -> str:
    return json.dumps(storage_config, sort_keys=True)


DEFAULT_TRANSFER_POOL = TransferPool()
with_transfer = DEFAULT_TRANSFER_POOL.with_transfer
get_transfer = DEFAULT_TRANSFER_POOL.get_transfer



# Copyright (c) 2021 Aiven, Helsinki, Finland. https://aiven.io/

from __future__ import annotations

from pathlib import Path
from rohmu.delta.common import (
    BackupPath,
    EMBEDDED_FILE_SIZE,
    hash_hexdigest_readable,
    increase_worth_reporting,
    parallel_map_to,
    Progress,
    SnapshotFile,
    SnapshotHash,
    SnapshotState,
)
from rohmu.typing import StrOrPathLike
from typing import Any, Callable, Generator, Iterable, List, Optional, Sequence, Set, Tuple, Union

import base64
import logging
import os
import threading

logger = logging.getLogger(__name__)


class Snapshotter:
    """Snapshotter keeps track of files on disk, and their hashes.

    The hash on disk MAY change, which may require subsequent
    incremental snapshot and-or ignoring the files which have changed.

    The output to outside is just root object's hash, as well as list
    of other hashes which correspond to files referred to within the
    file list contained in root object.

    Note that any call to public API MUST be made with
    snapshotter.lock held. This is because Snapshotter is process-wide
    utility that is shared across operations, possibly used from
    multiple threads, and the single-operation-only mode of operation
    is not exactly flawless (the 'new operation can be started with
    old running' is intentional feature but new operation should
    eventually replace the old). The lock itself might not need to be
    built-in to Snapshotter, but having it there enables asserting its
    state during public API calls.
    """

    def __init__(
        self,
        *,
        src: StrOrPathLike,
        dst: StrOrPathLike,
        globs: list[str],
        src_iterate_func: Optional[Callable[[], Iterable[Union[BackupPath, str, Path]]]] = None,
        parallel: int = 1,
        min_delta_file_size: int = 0,
    ) -> None:
        assert globs
        self.src = Path(src)
        self.dst = Path(dst)
        self.globs = globs
        self.src_iterate_func = src_iterate_func
        self.relative_path_to_snapshotfile: dict[Path, SnapshotFile] = {}
        self.hexdigest_to_snapshotfiles: dict[str, List[SnapshotFile]] = {}
        self.parallel = parallel
        self.lock = threading.Lock()
        self.empty_dirs: List[Path] = []
        self.min_delta_file_size = min_delta_file_size

    def _list_files(self, basepath: Path) -> List[Path]:
        result_files = set()
        for glob in self.globs:
            for path in basepath.glob(glob):
                if not path.is_file() or path.is_symlink():
                    continue
                relpath = path.relative_to(basepath)
                result_files.add(relpath)

        return sorted(result_files)

    def _list_dirs_and_files(self, basepath: Path) -> Tuple[List[Path], List[Path]]:
        files = self._list_files(basepath)
        dirs = {p.parent for p in files}
        return sorted(dirs), files

    def _add_snapshotfile(self, snapshotfile: SnapshotFile) -> None:
        old_snapshotfile = self.relative_path_to_snapshotfile.get(snapshotfile.relative_path, None)
        if old_snapshotfile:
            self._remove_snapshotfile(old_snapshotfile)
        self.relative_path_to_snapshotfile[snapshotfile.relative_path] = snapshotfile
        if snapshotfile.hexdigest:
            self.hexdigest_to_snapshotfiles.setdefault(snapshotfile.hexdigest, []).append(snapshotfile)

    def _remove_snapshotfile(self, snapshotfile: SnapshotFile) -> None:
        assert self.relative_path_to_snapshotfile[snapshotfile.relative_path] == snapshotfile
        del self.relative_path_to_snapshotfile[snapshotfile.relative_path]
        if snapshotfile.hexdigest:
            self.hexdigest_to_snapshotfiles[snapshotfile.hexdigest].remove(snapshotfile)

    def _snapshotfile_from_path(self, relative_path: Path, missing_ok: bool) -> SnapshotFile:
        src_path = self.src / relative_path
        st = src_path.stat()
        return SnapshotFile(
            relative_path=relative_path,
            mtime_ns=st.st_mtime_ns,
            file_size=st.st_size,
            stored_file_size=0,
            missing_ok=missing_ok,
        )

    def _gen_snapshot_hashes(
        self, relative_paths: Sequence[Path], reuse_old_snapshotfiles: bool, required_paths: Optional[Set[Path]] = None
    ) -> Generator[SnapshotFile, None, None]:
        same = 0
        lost = 0

        if required_paths:
            missing_files = required_paths.difference(set(relative_paths))
            if missing_files:
                logger.error("Required file(s) disappeared during the backup: %r", missing_files)
                raise FileNotFoundError("File(s) disappeared during the backup, aborting.")

        for relative_path in relative_paths:
            old_snapshotfile = self.relative_path_to_snapshotfile.get(relative_path)
            missing_ok = True
            if required_paths and relative_path in required_paths:
                missing_ok = False
            try:
                snapshotfile = self._snapshotfile_from_path(relative_path, missing_ok=missing_ok)
            except FileNotFoundError as e:
                if not missing_ok:
                    raise FileNotFoundError(f"Required file disappeared during the backup: {relative_path}") from e
                lost += 1
                if increase_worth_reporting(lost):
                    logger.debug("#%d. lost - %s disappeared before stat, ignoring", lost, self.src / relative_path)
                continue
            if reuse_old_snapshotfiles and old_snapshotfile:
                snapshotfile.hexdigest = old_snapshotfile.hexdigest
                snapshotfile.content_b64 = old_snapshotfile.content_b64
                snapshotfile.should_be_bundled = old_snapshotfile.should_be_bundled
                snapshotfile.missing_ok = old_snapshotfile.missing_ok
                if old_snapshotfile == snapshotfile:
                    same += 1
                    if increase_worth_reporting(same):
                        logger.debug("#%d. same - %r in %s is same", same, old_snapshotfile, relative_path)
                    continue
            yield snapshotfile

    def get_snapshot_hashes(self) -> List[SnapshotHash]:
        assert self.lock.locked()
        return [
            SnapshotHash(hexdigest=dig, size=sf[0].file_size) for dig, sf in self.hexdigest_to_snapshotfiles.items() if sf
        ]

    def get_snapshot_state(self) -> SnapshotState:
        assert self.lock.locked()
        return SnapshotState(
            root_globs=self.globs, files=sorted(self.relative_path_to_snapshotfile.values()), empty_dirs=self.empty_dirs
        )

    def update_snapshot_file_data(
        self, *, relative_path: Path, hexdigest: str, file_size: int, stored_file_size: int
    ) -> None:
        snapshotfile = self.relative_path_to_snapshotfile[relative_path]
        snapshotfile.hexdigest = hexdigest
        snapshotfile.file_size = file_size
        snapshotfile.stored_file_size = stored_file_size

    def _snapshot_create_missing_directories(self, *, src_dirs: Sequence[Path], dst_dirs: Sequence[Path]) -> int:
        changes = 0
        for i, relative_dir in enumerate(set(src_dirs).difference(dst_dirs), 1):
            dst_path = self.dst / relative_dir
            dst_path.mkdir(parents=True, exist_ok=True)
            if increase_worth_reporting(i):
                logger.debug("#%d. new directory: %r", i, relative_dir)
            changes += 1
        return changes

    def _snapshot_remove_extra_files(self, *, src_files: Sequence[Path], dst_files: Sequence[Path]) -> int:
        changes = 0
        for i, relative_path in enumerate(set(dst_files).difference(src_files), 1):
            dst_path = self.dst / relative_path
            snapshotfile = self.relative_path_to_snapshotfile.get(relative_path)
            if snapshotfile:
                self._remove_snapshotfile(snapshotfile)
            dst_path.unlink()
            if increase_worth_reporting(i):
                logger.debug("#%d. extra file: %r", i, relative_path)
            changes += 1
        return changes

    def _snapshot_add_missing_files(self, *, src_files: Sequence[Path], dst_files: Sequence[Path]) -> int:
        existing = 0
        disappeared = 0
        changes = 0
        for i, relative_path in enumerate(set(src_files).difference(dst_files), 1):
            src_path = self.src / relative_path
            dst_path = self.dst / relative_path
            try:
                os.link(src=src_path, dst=dst_path, follow_symlinks=False)
            except FileExistsError:
                # This happens only if snapshot is started twice at
                # same time. While it is technically speaking upstream
                # error, we rather handle it here than leave
                # exceptions not handled.
                existing += 1
                if increase_worth_reporting(existing):
                    logger.debug("#%d. %s already existed, ignoring", existing, src_path)
                continue
            except FileNotFoundError:
                disappeared += 1
                if increase_worth_reporting(disappeared):
                    logger.debug("#%d. %s disappeared before linking, ignoring", disappeared, src_path)
                continue
            if increase_worth_reporting(i - disappeared):
                logger.debug("#%d. new file: %r", i - disappeared, relative_path)
            changes += 1
        return changes

    def snapshot(self, *, progress: Optional[Progress] = None, reuse_old_snapshotfiles: bool = True) -> int:
        assert self.lock.locked()

        if progress is None:
            progress = Progress()
        progress.start(3)

        required_paths: Set[Path] = set()

        if self.src_iterate_func:
            src_dirs_set = set()
            src_files_set = set()
            for item in self.src_iterate_func():
                if isinstance(item, BackupPath):
                    missing_ok = item.missing_ok
                    path = item.path
                else:
                    # Default behaviour for backward compatibility
                    path = Path(item)
                    missing_ok = True

                if path.is_file() and not path.is_symlink():
                    if not missing_ok:
                        required_paths.add(path.relative_to(self.src))
                    src_files_set.add(path.relative_to(self.src))
                elif path.is_dir():
                    src_dirs_set.add(path.relative_to(self.src))
                elif not missing_ok:
                    raise FileNotFoundError(f"Required file disappeared during the backup: {path}")

            src_dirs = sorted(src_dirs_set | {p.parent for p in src_files_set})
            src_files = sorted(src_files_set)
        else:
            src_dirs, src_files = self._list_dirs_and_files(self.src)

        dst_dirs, dst_files = self._list_dirs_and_files(self.dst)

        # Create missing directories
        changes = self._snapshot_create_missing_directories(src_dirs=src_dirs, dst_dirs=dst_dirs)
        progress.add_success()

        # Remove extra files
        changes += self._snapshot_remove_extra_files(src_files=src_files, dst_files=dst_files)
        progress.add_success()

        # Add missing files
        changes += self._snapshot_add_missing_files(src_files=src_files, dst_files=dst_files)
        progress.add_success()

        # We COULD also remove extra directories, but it is not
        # probably really worth it and due to ignored files it
        # actually might not even work.

        # Then, create/update corresponding snapshotfile objects (old
        # ones were already removed)
        dst_dirs, dst_files = self._list_dirs_and_files(self.dst)
        self.empty_dirs = src_dirs
        snapshotfiles = list(self._gen_snapshot_hashes(dst_files, reuse_old_snapshotfiles, required_paths=required_paths))

        progress.add_total(len(snapshotfiles))

        def _cb(snapshotfile: SnapshotFile) -> SnapshotFile:
            # src may or may not be present; dst is present as it is in snapshot
            with snapshotfile.open_for_reading(self.dst) as f:
                if snapshotfile.file_size <= EMBEDDED_FILE_SIZE:
                    snapshotfile.content_b64 = base64.b64encode(f.read()).decode()
                elif snapshotfile.file_size < self.min_delta_file_size:
                    snapshotfile.should_be_bundled = True
                else:
                    snapshotfile.hexdigest = hash_hexdigest_readable(f)
            return snapshotfile

        def _result_cb(*, map_in: Any, map_out: SnapshotFile) -> bool:
            self._add_snapshotfile(map_out)
            assert progress is not None
            progress.add_success()
            return True

        changes += len(snapshotfiles)
        parallel_map_to(iterable=snapshotfiles, fun=_cb, result_callback=_result_cb, n=self.parallel)
        return changes



"""Copyright (c) 2022 Aiven, Helsinki, Finland. https://aiven.io/"""
from io import BytesIO
from rohmu.errors import InvalidByteRangeError
from rohmu.object_storage.base import KEY_TYPE_OBJECT
from rohmu.object_storage.local import LocalTransfer
from tempfile import NamedTemporaryFile, TemporaryDirectory
from unittest.mock import MagicMock

import json
import os
import pytest


def test_store_file_from_disk() -> None:
    with TemporaryDirectory() as destdir:
        notifier = MagicMock()
        transfer = LocalTransfer(
            directory=destdir,
            notifier=notifier,
        )
        test_data = b"test-data"
        with NamedTemporaryFile() as tmpfile:
            tmpfile.write(test_data)
            tmpfile.flush()
            transfer.store_file_from_disk(key="test_key1", filepath=tmpfile.name)

        assert open(os.path.join(destdir, "test_key1"), "rb").read() == test_data
        notifier.object_created.assert_called_once_with(
            key="test_key1", size=len(test_data), metadata={"Content-Length": "9"}
        )


def test_store_file_object() -> None:
    with TemporaryDirectory() as destdir:
        notifier = MagicMock()
        transfer = LocalTransfer(
            directory=destdir,
            notifier=notifier,
        )
        test_data = b"test-data-2"
        file_object = BytesIO(test_data)

        transfer.store_file_object(key="test_key2", fd=file_object)

        assert open(os.path.join(destdir, "test_key2"), "rb").read() == test_data
        notifier.object_created.assert_called_once_with(key="test_key2", size=len(test_data), metadata={})

        data, _ = transfer.get_contents_to_string("test_key2")
        assert data == test_data

        data, _ = transfer.get_contents_to_string("test_key2", byte_range=(1, 123456))
        assert data == test_data[1:]

        data, _ = transfer.get_contents_to_string("test_key2", byte_range=(0, len(test_data) - 2))
        assert data == test_data[:-1]


def test_get_contents_to_fileobj_raises_error_on_invalid_byte_range() -> None:
    with TemporaryDirectory() as destdir:
        notifier = MagicMock()
        transfer = LocalTransfer(
            directory=destdir,
            notifier=notifier,
        )
        with pytest.raises(InvalidByteRangeError):
            transfer.get_contents_to_fileobj(
                key="testkey",
                fileobj_to_store_to=BytesIO(),
                byte_range=(100, 10),
            )


def test_can_handle_metadata_without_md5() -> None:
    with TemporaryDirectory() as destdir:
        notifier = MagicMock()
        transfer = LocalTransfer(
            directory=destdir,
            notifier=notifier,
        )
        test_data = b"test-data"
        with NamedTemporaryFile() as tmpfile:
            tmpfile.write(test_data)
            tmpfile.flush()
            transfer.store_file_from_disk(key="test_key1", filepath=tmpfile.name)

        # override the metadata file removing the hash.
        # this simulates files stored by older versions of rohmu
        target_file = transfer.format_key_for_backend("test_key1")
        metadata_file = target_file + ".metadata"
        old_metadata = {"Content-Length": str(len(test_data))}
        with open(metadata_file, "w") as metadata_fp:
            json.dump(old_metadata, metadata_fp)

        # we can read the metadata
        assert transfer.get_metadata_for_key("test_key1") == old_metadata
        # and we can also load the file information iterating over the storage
        item = next(transfer.iter_key("test_key1", with_metadata=True, include_key=True))
        assert item.type == KEY_TYPE_OBJECT
        result = item.value
        assert isinstance(result, dict)
        expected_value = {
            "name": "test_key1",
            "size": len(test_data),
            "metadata": old_metadata,
        }
        last_modified = result.pop("last_modified")
        assert last_modified is not None
        assert result == expected_value



from rohmu import Config, get_class_for_transfer, get_transfer
from unittest.mock import MagicMock, Mock, patch

import pytest
import sys


@pytest.mark.parametrize(
    "config",
    [
        {
            "storage_type": "s3",
            "region": "dummy",
            "bucket_name": "dummy",
            "notifier": {"notifier_type": "http", "url": "localhost"},
        }
    ],
)
@patch("rohmu.notifier.http.BackgroundHTTPNotifier")
@patch("rohmu.object_storage.s3.S3Transfer.from_model")
@patch("rohmu.object_storage.s3.S3Transfer.config_model")
def test_get_transfer(mock_config_model: Mock, mock_from_model: Mock, mock_notifier: Mock, config: Config) -> None:
    _transfer = get_transfer(config)
    expected_config_arg = dict(config)
    expected_config_arg.pop("storage_type")
    expected_config_arg.pop("notifier")
    mock_config_model.assert_called_once_with(**expected_config_arg, notifier=mock_notifier())
    mock_from_model.assert_called_once_with(mock_config_model())


@pytest.mark.parametrize("storage_type", ["s3", "local", "azure", "google", "swift", "s3"])
@patch.dict(sys.modules, {"swiftclient": MagicMock(), "azure.common": MagicMock()})
def test_config_model_defined(storage_type: str) -> None:
    assert get_class_for_transfer({"storage_type": storage_type}).config_model
